import numpy as np
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet.meshes.harmonic_coefficients.mesh_field import MeshField
from sleplet.meshes.mesh_harmonic_coefficients import MeshHarmonicCoefficients
from sleplet.utils.noise import compute_snr, create_mesh_noise
from sleplet.utils.string_methods import filename_args
from sleplet.utils.validation import Validation


@dataclass(config=Validation, kw_only=True)
class MeshNoiseField(MeshHarmonicCoefficients):
    SNR: float = 10

    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        mf = MeshField(self.mesh)
        noise = create_mesh_noise(mf.coefficients, self.SNR)
        compute_snr(mf.coefficients, noise, "Harmonic")
        return noise

    def _create_name(self) -> str:
        return f"{self.mesh.name}_noise_field{filename_args(self.SNR, 'snr')}"

    def _set_reality(self) -> bool:
        return True

    def _set_spin(self) -> int:
        return 0

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 1
            if len(self.extra_args) != num_args:
                raise ValueError(f"The number of extra arguments should be {num_args}")
            self.SNR = self.extra_args[0]



from dataclasses import KW_ONLY

import numpy as np
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet import logger
from sleplet.meshes.classes.mesh_slepian import MeshSlepian
from sleplet.utils.harmonic_methods import mesh_inverse
from sleplet.utils.integration_methods import (
    integrate_region_mesh,
    integrate_whole_mesh,
)
from sleplet.utils.validation import Validation


@dataclass(config=Validation)
class MeshSlepianDecomposition:
    mesh_slepian: MeshSlepian
    _: KW_ONLY
    mask: bool = False
    u_i: npt.NDArray[np.complex_ | np.float_] | None = None
    u: npt.NDArray[np.complex_ | np.float_] | None = None

    def __post_init_post_parse__(self) -> None:
        """ """
        self._detect_method()

    def decompose(self, rank: int) -> float:
        """
        decompose the signal into its Slepian coefficients via the given method
        """
        self._validate_rank(rank)

        match self.method:
            case "harmonic_sum":
                return self._harmonic_sum(rank)
            case "integrate_mesh":
                return self._integrate_mesh(rank)
            case "integrate_region":
                return self._integrate_region(rank)
            case _:
                raise ValueError(f"'{self.method}' is not a valid method")

    def decompose_all(self, n_coefficients: int) -> npt.NDArray[np.float_]:
        """
        decompose all ranks of the Slepian coefficients
        """
        coefficients = np.zeros(n_coefficients)
        for rank in range(n_coefficients):
            coefficients[rank] = self.decompose(rank)
        return coefficients

    def _integrate_region(self, rank: int) -> float:
        r"""
        f_{p} =
        \frac{1}{\lambda_{p}}
        \int\limits_{R} \dd{x}
        f(x) \overline{S_{p}(x)}
        """
        assert isinstance(self.u, np.ndarray)  # noqa: S101
        s_p = mesh_inverse(
            self.mesh_slepian.mesh,
            self.mesh_slepian.slepian_functions[rank],
        )
        integration = integrate_region_mesh(
            self.mesh_slepian.mesh.region,
            self.mesh_slepian.mesh.vertices,
            self.mesh_slepian.mesh.faces,
            self.u,
            s_p,
        )
        return integration / self.mesh_slepian.slepian_eigenvalues[rank]

    def _integrate_mesh(self, rank: int) -> float:
        r"""
        f_{p} =
        \int\limits_{x} \dd{x}
        f(x) \overline{S_{p}(x)}
        """
        assert isinstance(self.u, np.ndarray)  # noqa: S101
        s_p = mesh_inverse(
            self.mesh_slepian.mesh,
            self.mesh_slepian.slepian_functions[rank],
        )
        return integrate_whole_mesh(
            self.mesh_slepian.mesh.vertices, self.mesh_slepian.mesh.faces, self.u, s_p
        )

    def _harmonic_sum(self, rank: int) -> float:
        r"""
        f_{p} =
        \sum\limits_{i=0}^{K}
        f_{i} (S_{p})_{i}^{*}
        """
        return (self.u_i * self.mesh_slepian.slepian_functions[rank]).sum()

    def _detect_method(self) -> None:
        """
        detects what method is used to perform the decomposition
        """
        if isinstance(self.u_i, np.ndarray):
            logger.info("harmonic sum method selected")
            self.method = "harmonic_sum"
        elif isinstance(self.u, np.ndarray) and not self.mask:
            logger.info("integrating the whole mesh method selected")
            self.method = "integrate_mesh"
        elif isinstance(self.u, np.ndarray):
            logger.info("integrating a region on the mesh method selected")
            self.method = "integrate_region"
        else:
            raise RuntimeError(
                "need to pass one off harmonic coefficients, real pixels "
                "or real pixels with a mask"
            )

    def _validate_rank(self, rank: int) -> None:
        """
        checks the requested rank is valid
        """
        assert isinstance(  # noqa: S101
            self.mesh_slepian.mesh.number_basis_functions, int
        )
        if not isinstance(rank, int):
            raise TypeError("rank should be an integer")
        if rank < 0:
            raise ValueError("rank cannot be negative")
        if rank >= self.mesh_slepian.mesh.number_basis_functions:
            raise ValueError(
                "rank should be less than "
                f"{self.mesh_slepian.mesh.number_basis_functions}"
            )



import numpy as np
import pyssht as ssht
from numpy.testing import assert_allclose, assert_raises

from sleplet.slepian.slepian_decomposition import SlepianDecomposition
from sleplet.utils.mask_methods import create_mask_region
from sleplet.utils.slepian_methods import slepian_forward, slepian_inverse
from sleplet.utils.vars import SAMPLING_SCHEME


def test_decompose_all_polar(slepian_polar_cap, earth_polar_cap) -> None:
    """
    tests that all three methods produce the same coefficients for polar cap
    """
    field = ssht.inverse(
        earth_polar_cap.coefficients, slepian_polar_cap.L, Method=SAMPLING_SCHEME
    )
    harmonic_sum_p = slepian_forward(
        slepian_polar_cap.L, slepian_polar_cap, flm=earth_polar_cap.coefficients
    )
    integrate_sphere_p = slepian_forward(
        slepian_polar_cap.L, slepian_polar_cap, f=field
    )
    integrate_region_p = slepian_forward(
        slepian_polar_cap.L, slepian_polar_cap, f=field, mask=slepian_polar_cap.mask
    )
    assert_allclose(
        np.abs(integrate_sphere_p - harmonic_sum_p)[: slepian_polar_cap.N].mean(),
        0,
        atol=10,
    )
    assert_allclose(
        np.abs(integrate_region_p - harmonic_sum_p)[: slepian_polar_cap.N].mean(),
        0,
        atol=17,
    )


def test_decompose_all_lim_lat_lon(slepian_lim_lat_lon, earth_lim_lat_lon) -> None:
    """
    tests that all three methods produce the same coefficients for
    limited latitude longitude region
    """
    field = ssht.inverse(
        earth_lim_lat_lon.coefficients, slepian_lim_lat_lon.L, Method=SAMPLING_SCHEME
    )
    harmonic_sum_p = slepian_forward(
        slepian_lim_lat_lon.L, slepian_lim_lat_lon, flm=earth_lim_lat_lon.coefficients
    )
    integrate_sphere_p = slepian_forward(
        slepian_lim_lat_lon.L, slepian_lim_lat_lon, f=field
    )
    integrate_region_p = slepian_forward(
        slepian_lim_lat_lon.L,
        slepian_lim_lat_lon,
        f=field,
        mask=slepian_lim_lat_lon.mask,
    )
    assert_allclose(
        np.abs(integrate_sphere_p - harmonic_sum_p)[: slepian_lim_lat_lon.N].mean(),
        0,
        atol=3.5,
    )
    assert_allclose(
        np.abs(integrate_region_p - harmonic_sum_p)[: slepian_lim_lat_lon.N].mean(),
        0,
        atol=90,
    )


def test_equality_to_harmonic_transform_polar(
    slepian_polar_cap, earth_polar_cap
) -> None:
    """
    tests that fp*Sp up to N is roughly equal to flm*Ylm
    """
    f_p = slepian_forward(
        slepian_polar_cap.L, slepian_polar_cap, flm=earth_polar_cap.coefficients
    )
    f_slepian = slepian_inverse(f_p, slepian_polar_cap.L, slepian_polar_cap)
    f_harmonic = ssht.inverse(
        earth_polar_cap.coefficients, slepian_polar_cap.L, Method=SAMPLING_SCHEME
    )
    mask = create_mask_region(slepian_polar_cap.L, slepian_polar_cap.region)
    assert_allclose(np.abs(f_slepian - f_harmonic)[mask].mean(), 0, atol=89)


def test_equality_to_harmonic_transform_lim_lat_lon(
    slepian_lim_lat_lon, earth_lim_lat_lon
) -> None:
    """
    tests that fp*Sp up to N is roughly equal to flm*Ylm
    """
    f_p = slepian_forward(
        slepian_lim_lat_lon.L, slepian_lim_lat_lon, flm=earth_lim_lat_lon.coefficients
    )
    f_slepian = slepian_inverse(f_p, slepian_lim_lat_lon.L, slepian_lim_lat_lon)
    f_harmonic = ssht.inverse(
        earth_lim_lat_lon.coefficients, slepian_lim_lat_lon.L, Method=SAMPLING_SCHEME
    )
    mask = create_mask_region(slepian_lim_lat_lon.L, slepian_lim_lat_lon.region)
    assert_allclose(np.abs(f_slepian - f_harmonic)[mask].mean(), 0, atol=248)


def test_pass_rank_higher_than_available(slepian_polar_cap, earth_polar_cap) -> None:
    """
    tests that asking for a Slepian coefficients above the limit fails
    """
    sd = SlepianDecomposition(
        slepian_polar_cap.L, slepian_polar_cap, flm=earth_polar_cap.coefficients
    )
    assert_raises(ValueError, sd.decompose, slepian_polar_cap.L**2)


def test_no_method_found_for_decomposition(slepian_polar_cap) -> None:
    """
    checks that no method has been found when inputs haven't been set
    """
    assert_raises(
        RuntimeError, SlepianDecomposition, slepian_polar_cap.L, slepian_polar_cap
    )



from abc import abstractmethod

import numpy as np
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet.meshes.mesh_coefficients import MeshCoefficients
from sleplet.utils.noise import compute_snr, create_mesh_noise
from sleplet.utils.validation import Validation


@dataclass(config=Validation)
class MeshHarmonicCoefficients(MeshCoefficients):
    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _add_noise_to_signal(
        self,
    ) -> tuple[npt.NDArray[np.complex_ | np.float_] | None, float | None]:
        """
        adds Gaussian white noise to the signal
        """
        self.coefficients: npt.NDArray[np.complex_ | np.float_]
        if self.noise is not None:
            unnoised_coefficients = self.coefficients.copy()
            nlm = create_mesh_noise(self.coefficients, self.noise)
            snr = compute_snr(self.coefficients, nlm, "Harmonic")
            self.coefficients = self.coefficients + nlm
            return unnoised_coefficients, snr
        return None, None

    @abstractmethod
    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        raise NotImplementedError

    @abstractmethod
    def _create_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def _setup_args(self) -> None:
        raise NotImplementedError



from pathlib import Path

from sleplet.data.setup_pooch import find_on_pooch_then_local


def test_known_file_exists_on_pooch() -> None:
    """
    checks that a known file exists on the Pooch repository
    """
    filename = "slepian_masks_south_america_L1.npy"
    assert isinstance(find_on_pooch_then_local(filename), str)


def test_known_file_exists_in_data_folder() -> None:
    """
    checks that a known file exists in the data folder
    """
    filename = "meshes_regions_bird.toml"
    assert isinstance(find_on_pooch_then_local(filename), Path)


def test_unknown_file_returns_none() -> None:
    """
    checks made up file doesn't exist
    """
    filename = "xyz"
    assert find_on_pooch_then_local(filename) is None



import numpy as np
import pytest
from numpy.testing import assert_allclose

from sleplet.slepian.slepian_region.slepian_arbitrary import SlepianArbitrary


@pytest.mark.slow
def test_equality_to_polar_cap_method(slepian_polar_cap) -> None:
    """
    tests that the eigenvectors and eigenvalues are close
    in comparison to the smarter Slepian polar cap method
    """
    mask_name = slepian_polar_cap.region.name_ending
    slepian = SlepianArbitrary(slepian_polar_cap.L, mask_name)
    assert_allclose(np.abs(slepian_polar_cap.N - slepian.N), 0, atol=3)
    assert_allclose(
        np.abs(
            slepian.eigenvalues[: slepian_polar_cap.N]
            - slepian_polar_cap.eigenvalues[: slepian_polar_cap.N]
        ).mean(),
        0,
        atol=0.88,
    )
    assert_allclose(
        np.abs(
            slepian.eigenvectors[: slepian_polar_cap.N]
            - slepian_polar_cap.eigenvectors[: slepian_polar_cap.N]
        ).mean(),
        0,
        atol=0.03,
    )


@pytest.mark.slow
def test_equality_to_lim_lat_lon_method(slepian_lim_lat_lon) -> None:
    """
    tests that the eigenvectors and eigenvalues are close
    in comparison to the smarter Slepian lim lat lon method
    """
    mask_name = slepian_lim_lat_lon.region.name_ending
    slepian = SlepianArbitrary(slepian_lim_lat_lon.L, mask_name)
    assert_allclose(np.abs(slepian_lim_lat_lon.N - slepian.N), 0, atol=1)
    assert_allclose(
        np.abs(
            slepian.eigenvalues[: slepian.N]
            - slepian_lim_lat_lon.eigenvalues[: slepian.N]
        ).mean(),
        0,
        atol=0.03,
    )
    assert_allclose(
        np.abs(slepian.eigenvectors - slepian.eigenvectors)[: slepian.N].mean(),
        0,
        atol=0.0,
    )



import numpy as np
import pyssht as ssht
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet.data.other.earth.create_earth_flm import create_flm
from sleplet.data.setup_pooch import find_on_pooch_then_local
from sleplet.functions.f_lm import F_LM
from sleplet.utils.harmonic_methods import (
    ensure_f_bandlimited,
    rotate_earth_to_south_america,
)
from sleplet.utils.mask_methods import create_mask
from sleplet.utils.string_methods import convert_camel_case_to_snake_case
from sleplet.utils.validation import Validation
from sleplet.utils.vars import SAMPLING_SCHEME


@dataclass(config=Validation)
class SouthAmerica(F_LM):
    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        return ensure_f_bandlimited(
            self._grid_fun, self.L, reality=self.reality, spin=self.spin
        )

    def _create_name(self) -> str:
        return convert_camel_case_to_snake_case(self.__class__.__name__)

    def _set_reality(self) -> bool:
        return True

    def _set_spin(self) -> int:
        return 0

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            raise AttributeError(
                f"{self.__class__.__name__} does not support extra arguments"
            )

    def _grid_fun(
        self, theta: npt.NDArray[np.float_], phi: npt.NDArray[np.float_]
    ) -> npt.NDArray[np.float_]:
        """
        function on the grid
        """
        earth_flm = create_flm(self.L, smoothing=self.smoothing)
        rot_flm = rotate_earth_to_south_america(earth_flm, self.L)
        earth_f = ssht.inverse(
            rot_flm, self.L, Reality=self.reality, Method=SAMPLING_SCHEME
        )
        mask_name = f"{self.name}_L{self.L}.npy"
        mask_location = find_on_pooch_then_local(f"slepian_masks_{mask_name}")
        mask = (
            create_mask(self.L, mask_name)
            if mask_location is None
            else np.load(mask_location)
        )
        return np.where(mask, earth_f, 0)



from numpy.testing import assert_array_equal, assert_raises

from sleplet.functions.flm.earth import Earth
from sleplet.utils.vars import SMOOTHING

L = 128


def test_smoothing_changes_flm() -> None:
    """
    tests the addition of Gaussian noise changes the coefficients
    """
    earth = Earth(L)
    earth_smoothed = Earth(L, smoothing=SMOOTHING)
    assert_raises(
        AssertionError,
        assert_array_equal,
        earth.coefficients,
        earth_smoothed.coefficients,
    )



from abc import abstractmethod

import numpy as np
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet import logger
from sleplet.utils.region import Region
from sleplet.utils.validation import Validation


@dataclass(config=Validation)
class SlepianFunctions:
    L: int

    def __post_init_post_parse__(self) -> None:
        self.region = self._create_region()
        self.mask = self._create_mask()
        self.name = self._create_fn_name()
        area = self._calculate_area()
        self.N = round(area * self.L**2 / (4 * np.pi))
        logger.info(f"Shannon number N={self.N}")
        self.matrix_location = self._create_matrix_location()
        logger.info("start solving eigenproblem")
        self.eigenvalues, self.eigenvectors = self._solve_eigenproblem()
        logger.info("finished solving eigenproblem")

    @abstractmethod
    def _create_fn_name(self) -> str:
        """
        creates the name for plotting
        """
        raise NotImplementedError

    @abstractmethod
    def _create_region(self) -> Region:
        """
        creates a region object for area of interest
        """
        raise NotImplementedError

    @abstractmethod
    def _create_mask(self) -> npt.NDArray[np.float_]:
        """
        creates a mask of the region of interest
        """
        raise NotImplementedError

    @abstractmethod
    def _calculate_area(self) -> float:
        """
        calculates area of region
        """
        raise NotImplementedError

    @abstractmethod
    def _create_matrix_location(self) -> str:
        """
        creates the name of the matrix binary
        """
        raise NotImplementedError

    @abstractmethod
    def _solve_eigenproblem(
        self,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        """
        solves the eigenproblem for the given function
        """
        raise NotImplementedError



import numpy as np
from numpy import typing as npt
from pydantic import validator
from pydantic.dataclasses import dataclass
from pys2let import pys2let_j_max

from sleplet import logger
from sleplet.meshes.mesh_slepian_coefficients import MeshSlepianCoefficients
from sleplet.meshes.slepian_coefficients.mesh_slepian_field import MeshSlepianField
from sleplet.meshes.slepian_coefficients.mesh_slepian_wavelets import (
    MeshSlepianWavelets,
)
from sleplet.utils.string_methods import filename_args, wavelet_ending
from sleplet.utils.validation import Validation
from sleplet.utils.wavelet_methods import slepian_wavelet_forward


@dataclass(config=Validation, kw_only=True)
class MeshSlepianWaveletCoefficients(MeshSlepianCoefficients):
    B: int = 3
    j_min: int = 2
    j: int | None = None

    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        logger.info("start computing wavelet coefficients")
        self.wavelets, self.wavelet_coefficients = self._create_wavelet_coefficients()
        logger.info("finish computing wavelet coefficients")
        jth = 0 if self.j is None else self.j + 1
        return self.wavelet_coefficients[jth]

    def _create_name(self) -> str:
        return (
            f"slepian_wavelet_coefficients_{self.mesh.name}"
            f"{filename_args(self.B, 'B')}"
            f"{filename_args(self.j_min, 'jmin')}"
            f"{wavelet_ending(self.j_min, self.j)}"
        )

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 3
            if len(self.extra_args) != num_args:
                raise ValueError(f"The number of extra arguments should be {num_args}")
            self.B, self.j_min, self.j = self.extra_args

    def _create_wavelet_coefficients(
        self,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_ | np.float_]]:
        """
        computes wavelet coefficients in Slepian space
        """
        smw = MeshSlepianWavelets(self.mesh, B=self.B, j_min=self.j_min)
        smf = MeshSlepianField(self.mesh)
        wavelets = smw.wavelets
        wavelet_coefficients = slepian_wavelet_forward(
            smf.coefficients,
            wavelets,
            self.mesh_slepian.N,
        )
        return wavelets, wavelet_coefficients

    @validator("j")
    def check_j(cls, v, values):
        j_max = pys2let_j_max(
            values["B"], values["mesh"].mesh_eigenvalues.shape[0], values["j_min"]
        )
        if v is not None and v < 0:
            raise ValueError("j should be positive")
        if v is not None and v > j_max - values["j_min"]:
            raise ValueError(
                f"j should be less than j_max - j_min: {j_max - values['j_min'] + 1}"
            )
        return v



import numpy as np
from numpy.testing import assert_array_equal

from sleplet.utils.array_methods import fill_upper_triangle_of_hermitian_matrix


def test_fill_matrix_using_hermitian_relation() -> None:
    """
    test that Hermitian symmetry is applied to matrix
    """
    matrix_in = np.array(
        [[1 + 1j, 0, 0], [4 + 4j, 5 + 5j, 0], [7 + 7j, 8 + 8j, 9 + 9j]]
    )
    matrix_out = np.array(
        [[1 + 1j, 4 - 4j, 7 - 7j], [4 + 4j, 5 + 5j, 8 - 8j], [7 + 7j, 8 + 8j, 9 + 9j]]
    )
    fill_upper_triangle_of_hermitian_matrix(matrix_in)
    assert_array_equal(matrix_in, matrix_out)



import numpy as np
from numpy.testing import assert_equal

from sleplet.utils.region import Region

MASK = "south_america"
PHI_0 = np.pi / 6
PHI_1 = np.pi
THETA_0 = np.pi / 6
THETA_1 = np.pi / 3
THETA_MAX = 2 * np.pi / 9


def test_polar_cap_region() -> None:
    """
    tests that a polar cap is created
    """
    region = Region(theta_max=THETA_MAX)
    assert_equal(region.region_type, "polar")


def test_lim_lat_lon_region() -> None:
    """
    tests that a limited latitude longitude region is created
    """
    region = Region(theta_min=THETA_0, theta_max=THETA_1, phi_min=PHI_0, phi_max=PHI_1)
    assert_equal(region.region_type, "lim_lat_lon")


def test_arbitrary_region() -> None:
    """
    tests that an arbitrary region is created
    """
    region = Region(mask_name=MASK)
    assert_equal(region.region_type, "arbitrary")


def test_polar_hierarchy_over_arbirary() -> None:
    """
    ensures polar cap is made instead of arbitrary
    """
    region = Region(theta_max=THETA_MAX, mask_name=MASK)
    assert_equal(region.region_type, "polar")


def test_lim_lat_lon_hierarchy_over_arbirary() -> None:
    """
    ensures limited latitude longitude region is made instead of arbitrary
    """
    region = Region(
        theta_min=THETA_0,
        theta_max=THETA_1,
        phi_min=PHI_0,
        phi_max=PHI_1,
        mask_name=MASK,
    )
    assert_equal(region.region_type, "lim_lat_lon")



import numpy as np
import pyssht as ssht
from numpy import typing as npt
from numpy.random import default_rng

from sleplet import logger
from sleplet.meshes.classes.mesh_slepian import MeshSlepian
from sleplet.slepian.slepian_functions import SlepianFunctions
from sleplet.utils.harmonic_methods import mesh_forward
from sleplet.utils.slepian_methods import (
    compute_mesh_s_p_pixel,
    compute_s_p_omega,
    slepian_forward,
    slepian_inverse,
    slepian_mesh_forward,
    slepian_mesh_inverse,
)
from sleplet.utils.vars import RANDOM_SEED, SAMPLING_SCHEME


def _signal_power(signal: npt.NDArray[np.complex_ | np.float_]) -> float:
    """
    computes the power of the signal
    """
    return (np.abs(signal) ** 2).sum()


def compute_snr(
    signal: npt.NDArray[np.complex_ | np.float_],
    noise: npt.NDArray[np.complex_ | np.float_],
    signal_type: str,
) -> float:
    """
    computes the signal to noise ratio
    """
    snr = 10 * np.log10(_signal_power(signal) / _signal_power(noise))
    logger.info(f"{signal_type} SNR: {snr:.2f}")
    return snr


def compute_sigma_noise(
    signal: npt.NDArray[np.complex_ | np.float_],
    snr_in: float,
    *,
    denominator: int | None = None,
) -> float:
    """
    compute the std dev of the noise
    """
    if denominator is None:
        denominator = signal.shape[0]
    return np.sqrt(10 ** (-snr_in / 10) * _signal_power(signal) / denominator)


def create_noise(
    L: int, signal: npt.NDArray[np.complex_ | np.float_], snr_in: float
) -> npt.NDArray[np.complex_]:
    """
    computes Gaussian white noise
    """
    # set random seed
    rng = default_rng(RANDOM_SEED)

    # initialise
    nlm = np.zeros(L**2, dtype=np.complex_)

    # std dev of the noise
    sigma_noise = compute_sigma_noise(signal, snr_in)

    # compute noise
    for ell in range(L):
        ind = ssht.elm2ind(ell, 0)
        nlm[ind] = sigma_noise * rng.standard_normal()
        for m in range(1, ell + 1):
            ind_pm = ssht.elm2ind(ell, m)
            ind_nm = ssht.elm2ind(ell, -m)
            nlm[ind_pm] = (
                sigma_noise
                / np.sqrt(2)
                * (rng.standard_normal() + 1j * rng.standard_normal())
            )
            nlm[ind_nm] = (-1) ** m * nlm[ind_pm].conj()
    return nlm


def create_slepian_noise(
    L: int,
    slepian_signal: npt.NDArray[np.complex_ | np.float_],
    slepian: SlepianFunctions,
    snr_in: float,
) -> npt.NDArray[np.complex_]:
    """
    computes Gaussian white noise in Slepian space
    """
    flm = ssht.forward(
        slepian_inverse(slepian_signal, L, slepian), L, Method=SAMPLING_SCHEME
    )
    nlm = create_noise(L, flm, snr_in)
    return slepian_forward(L, slepian, flm=nlm)


def _perform_hard_thresholding(
    f: npt.NDArray[np.complex_ | np.float_],
    sigma_j: float | npt.NDArray[np.float_],
    n_sigma: int,
) -> npt.NDArray[np.complex_]:
    """
    set pixels in real space to zero if the magnitude is less than the threshold
    """
    threshold = n_sigma * sigma_j
    return np.where(np.abs(f) < threshold, 0, f)


def harmonic_hard_thresholding(
    L: int,
    wav_coeffs: npt.NDArray[np.complex_],
    sigma_j: npt.NDArray[np.float_],
    n_sigma: int,
) -> npt.NDArray[np.complex_]:
    """
    perform thresholding in harmonic space
    """
    logger.info("begin harmonic hard thresholding")
    for j, coefficient in enumerate(wav_coeffs[1:]):
        logger.info(f"start Psi^{j + 1}/{len(wav_coeffs)-1}")
        f = ssht.inverse(coefficient, L, Method=SAMPLING_SCHEME)
        f_thresholded = _perform_hard_thresholding(f, sigma_j[j], n_sigma)
        wav_coeffs[j + 1] = ssht.forward(f_thresholded, L, Method=SAMPLING_SCHEME)
    return wav_coeffs


def slepian_wavelet_hard_thresholding(
    L: int,
    wav_coeffs: npt.NDArray[np.complex_ | np.float_],
    sigma_j: npt.NDArray[np.float_],
    n_sigma: int,
    slepian: SlepianFunctions,
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    perform thresholding in Slepian wavelet space
    """
    logger.info("begin Slepian hard thresholding")
    for j, coefficient in enumerate(wav_coeffs):
        logger.info(f"start Psi^{j + 1}/{len(wav_coeffs)}")
        f = slepian_inverse(coefficient, L, slepian)
        f_thresholded = _perform_hard_thresholding(f, sigma_j[j], n_sigma)
        wav_coeffs[j] = slepian_forward(L, slepian, f=f_thresholded)
    return wav_coeffs


def slepian_function_hard_thresholding(
    L: int,
    coefficients: npt.NDArray[np.complex_ | np.float_],
    sigma: float,
    n_sigma: int,
    slepian: SlepianFunctions,
) -> npt.NDArray[np.complex_]:
    """
    perform thresholding in Slepian function space
    """
    logger.info("begin Slepian hard thresholding")
    f = slepian_inverse(coefficients, L, slepian)
    f_thresholded = _perform_hard_thresholding(f, sigma, n_sigma)
    return slepian_forward(L, slepian, f=f_thresholded)


def compute_sigma_j(
    signal: npt.NDArray[np.complex_ | np.float_],
    psi_j: npt.NDArray[np.complex_],
    snr_in: float,
) -> npt.NDArray[np.float_]:
    """
    compute sigma_j for wavelets used in denoising the signal
    """
    sigma_noise = compute_sigma_noise(signal, snr_in)
    wavelet_power = (np.abs(psi_j) ** 2).sum(axis=1)
    return sigma_noise * np.sqrt(wavelet_power)


def compute_slepian_sigma_j(
    L: int,
    signal: npt.NDArray[np.complex_ | np.float_],
    psi_j: npt.NDArray[np.float_],
    snr_in: float,
    slepian: SlepianFunctions,
) -> npt.NDArray[np.float_]:
    """
    compute sigma_j for wavelets used in denoising the signal
    """
    sigma_noise = compute_sigma_noise(signal, snr_in, denominator=L**2)
    s_p = compute_s_p_omega(L, slepian)
    psi_j_reshape = psi_j[:, : slepian.N, np.newaxis, np.newaxis]
    wavelet_power = (np.abs(psi_j_reshape) ** 2 * np.abs(s_p) ** 2).sum(axis=1)
    return sigma_noise * np.sqrt(wavelet_power)


def create_mesh_noise(
    u_i: npt.NDArray[np.complex_ | np.float_], snr_in: float
) -> npt.NDArray[np.float_]:
    """
    computes Gaussian white noise
    """
    # set random seed
    rng = default_rng(RANDOM_SEED)

    # initialise
    n_i = np.zeros(u_i.shape[0])

    # std dev of the noise
    sigma_noise = compute_sigma_noise(u_i, snr_in)

    # compute noise
    for i in range(u_i.shape[0]):
        n_i[i] = sigma_noise * rng.standard_normal()
    return n_i


def create_slepian_mesh_noise(
    mesh_slepian: MeshSlepian,
    slepian_signal: npt.NDArray[np.complex_ | np.float_],
    snr_in: float,
) -> npt.NDArray[np.float_]:
    """
    computes Gaussian white noise in Slepian space
    """
    u_i = mesh_forward(
        mesh_slepian.mesh,
        slepian_mesh_inverse(
            mesh_slepian,
            slepian_signal,
        ),
    )
    n_i = create_mesh_noise(u_i, snr_in)
    return slepian_mesh_forward(
        mesh_slepian,
        u_i=n_i,
    )


def compute_slepian_mesh_sigma_j(
    mesh_slepian: MeshSlepian,
    signal: npt.NDArray[np.complex_ | np.float_],
    psi_j: npt.NDArray[np.float_],
    snr_in: float,
) -> npt.NDArray[np.float_]:
    """
    compute sigma_j for wavelets used in denoising the signal
    """
    sigma_noise = compute_sigma_noise(
        signal, snr_in, denominator=mesh_slepian.slepian_eigenvalues.shape[0]
    )
    s_p = compute_mesh_s_p_pixel(mesh_slepian)
    psi_j_reshape = psi_j[:, : mesh_slepian.N, np.newaxis]
    wavelet_power = (np.abs(psi_j_reshape) ** 2 * np.abs(s_p) ** 2).sum(axis=1)
    return sigma_noise * np.sqrt(wavelet_power)


def slepian_mesh_hard_thresholding(
    mesh_slepian: MeshSlepian,
    wav_coeffs: npt.NDArray[np.complex_ | np.float_],
    sigma_j: npt.NDArray[np.float_],
    n_sigma: int,
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    perform thresholding in Slepian space
    """
    logger.info("begin Slepian mesh hard thresholding")
    for j, coefficient in enumerate(wav_coeffs):
        logger.info(f"start Psi^{j + 1}/{len(wav_coeffs)}")
        f = slepian_mesh_inverse(mesh_slepian, coefficient)
        f_thresholded = _perform_hard_thresholding(f, sigma_j[j], n_sigma)
        wav_coeffs[j] = slepian_mesh_forward(mesh_slepian, u=f_thresholded)
    return wav_coeffs



from numpy.testing import assert_allclose, assert_equal

from sleplet.utils.parallel_methods import split_arr_into_chunks

L_LARGE = 128
L_SMALL = 16
NCPU = 4


def test_split_L_into_chunks() -> None:  # noqa: N802
    """
    ensure vector L split into appropriate number of chunks
    """
    chunks = split_arr_into_chunks(L_SMALL, NCPU)
    assert_equal(len(chunks), NCPU)
    chunk_length = L_SMALL // NCPU
    for chunk in chunks:
        assert_allclose(len(chunk), chunk_length, atol=0)



import numpy as np
import pytest
from numpy import typing as npt
from numpy.random import default_rng

from sleplet.functions.flm.earth import Earth
from sleplet.functions.flm.south_america import SouthAmerica
from sleplet.functions.fp.slepian_dirac_delta import SlepianDiracDelta
from sleplet.functions.fp.slepian_wavelets import SlepianWavelets
from sleplet.meshes.classes.mesh import Mesh
from sleplet.meshes.classes.mesh_slepian import MeshSlepian
from sleplet.meshes.harmonic_coefficients.mesh_field import MeshField
from sleplet.meshes.slepian_coefficients.mesh_slepian_wavelets import (
    MeshSlepianWavelets,
)
from sleplet.slepian.slepian_region.slepian_arbitrary import SlepianArbitrary
from sleplet.slepian.slepian_region.slepian_limit_lat_lon import SlepianLimitLatLon
from sleplet.slepian.slepian_region.slepian_polar_cap import SlepianPolarCap
from sleplet.utils.harmonic_methods import compute_random_signal
from sleplet.utils.vars import RANDOM_SEED

ARRAY_DIM = 10
L = 16
MASK = "south_america"
PHI_0 = np.pi / 6
PHI_1 = np.pi / 3
THETA_0 = np.pi / 6
THETA_1 = np.pi / 3
THETA_MAX = 2 * np.pi / 9


@pytest.fixture(scope="session")
def slepian_polar_cap() -> SlepianPolarCap:
    """
    create a Slepian polar cap class
    """
    return SlepianPolarCap(L, THETA_MAX)


@pytest.fixture(scope="session")
def slepian_lim_lat_lon() -> SlepianLimitLatLon:
    """
    create a Slepian limited latitude longitude class
    """
    return SlepianLimitLatLon(
        L, theta_min=THETA_0, theta_max=THETA_1, phi_min=PHI_0, phi_max=PHI_1
    )


@pytest.fixture(scope="session")
def slepian_arbitrary() -> SlepianArbitrary:
    """
    create a Slepian arbitrary class
    """
    return SlepianArbitrary(L, MASK)


@pytest.fixture(scope="session")
def earth_polar_cap(slepian_polar_cap) -> Earth:
    """
    Earth with polar cap region
    """
    return Earth(slepian_polar_cap.L, region=slepian_polar_cap.region)


@pytest.fixture(scope="session")
def earth_lim_lat_lon(slepian_lim_lat_lon) -> Earth:
    """
    Earth with limited latitude longitude region
    """
    return Earth(slepian_lim_lat_lon.L, region=slepian_lim_lat_lon.region)


@pytest.fixture(scope="session")
def south_america_arbitrary(slepian_arbitrary) -> SouthAmerica:
    """
    South America already has region
    """
    return SouthAmerica(slepian_arbitrary.L)


@pytest.fixture(scope="session")
def slepian_dirac_delta_polar_cap(slepian_polar_cap) -> SlepianDiracDelta:
    """
    Creates a polar cap Slepian Dirac delta
    """
    return SlepianDiracDelta(slepian_polar_cap.L, region=slepian_polar_cap.region)


@pytest.fixture(scope="session")
def slepian_dirac_delta_lim_lat_lon(slepian_lim_lat_lon) -> SlepianDiracDelta:
    """
    Creates a limited latitude longitude Slepian Dirac delta
    """
    return SlepianDiracDelta(slepian_lim_lat_lon.L, region=slepian_lim_lat_lon.region)


@pytest.fixture(scope="session")
def slepian_wavelets_polar_cap(slepian_polar_cap) -> SlepianWavelets:
    """
    computes the Slepian wavelets for the polar cap region
    """
    return SlepianWavelets(slepian_polar_cap.L, region=slepian_polar_cap.region)


@pytest.fixture(scope="session")
def slepian_wavelets_lim_lat_lon(slepian_lim_lat_lon) -> SlepianWavelets:
    """
    computes the Slepian wavelets for the lim_lat_lon region
    """
    return SlepianWavelets(slepian_lim_lat_lon.L, region=slepian_lim_lat_lon.region)


@pytest.fixture(scope="session")
def random_flm() -> npt.NDArray[np.complex_]:
    """
    creates random flm
    """
    rng = default_rng(RANDOM_SEED)
    return compute_random_signal(L, rng, var_signal=1)


@pytest.fixture(scope="session")
def random_nd_flm() -> npt.NDArray[np.complex_]:
    """
    creates multiple random flm
    """
    rng = default_rng(RANDOM_SEED)
    return np.array(
        [compute_random_signal(L, rng, var_signal=1) for _ in range(ARRAY_DIM)]
    )


@pytest.fixture(scope="session")
def mesh() -> Mesh:
    """
    creates a bird mesh
    """
    return Mesh("bird")


@pytest.fixture(scope="session")
def mesh_slepian(mesh) -> MeshSlepian:
    """
    creates a Slepian bird mesh
    """
    return MeshSlepian(mesh)


@pytest.fixture(scope="session")
def mesh_field_region(mesh) -> MeshField:
    """
    creates a field on the mesh
    """
    return MeshField(mesh, region=True)


@pytest.fixture(scope="session")
def mesh_slepian_wavelets(mesh) -> MeshSlepianWavelets:
    """
    creates a field on the mesh
    """
    return MeshSlepianWavelets(mesh)



from pydantic import validator
from pydantic.dataclasses import dataclass

from sleplet.utils.bool_methods import is_limited_lat_lon, is_polar_cap
from sleplet.utils.string_methods import angle_as_degree, multiples_of_pi
from sleplet.utils.validation import Validation
from sleplet.utils.vars import (
    PHI_MAX_DEFAULT,
    PHI_MIN_DEFAULT,
    THETA_MAX_DEFAULT,
    THETA_MIN_DEFAULT,
)


@dataclass(config=Validation, kw_only=True)
class Region:
    gap: bool = False
    mask_name: str = ""
    phi_max: float = PHI_MAX_DEFAULT
    phi_min: float = PHI_MIN_DEFAULT
    theta_max: float = THETA_MAX_DEFAULT
    theta_min: float = THETA_MIN_DEFAULT

    def __post_init_post_parse__(self) -> None:
        self._identify_region()

    def _identify_region(self) -> None:
        """
        identify region type based on the angle inputs or a mask name
        """
        if is_polar_cap(self.phi_min, self.phi_max, self.theta_min, self.theta_max):
            self.region_type = "polar"
            self.name_ending = (
                f"polar{'_gap' if self.gap else ''}"
                f"{angle_as_degree(self.theta_max)}"
            )

        elif is_limited_lat_lon(
            self.phi_min, self.phi_max, self.theta_min, self.theta_max
        ):
            self.region_type = "lim_lat_lon"
            self.name_ending = (
                f"theta{angle_as_degree(self.theta_min)}"
                f"-{angle_as_degree(self.theta_max)}"
                f"_phi{angle_as_degree(self.phi_min)}"
                f"-{angle_as_degree(self.phi_max)}"
            )

        elif self.mask_name:
            self.region_type = "arbitrary"
            self.name_ending = self.mask_name

        else:
            raise AttributeError(
                "need to specify either a polar cap, a limited latitude "
                "longitude region, or a file with a mask"
            )

    @validator("phi_max")
    def check_phi_max(cls, v):
        if v < PHI_MIN_DEFAULT:
            raise ValueError("phi_max cannot be negative")
        if v > PHI_MAX_DEFAULT:
            raise ValueError(
                f"phi_max cannot be greater than {multiples_of_pi(PHI_MAX_DEFAULT)}"
            )
        return v

    @validator("phi_min")
    def check_phi_min(cls, v):
        if v < PHI_MIN_DEFAULT:
            raise ValueError("phi_min cannot be negative")
        if v > PHI_MAX_DEFAULT:
            raise ValueError(
                f"phi_min cannot be greater than {multiples_of_pi(PHI_MAX_DEFAULT)}"
            )
        return v

    @validator("theta_max")
    def check_theta_max(cls, v):
        if v < THETA_MIN_DEFAULT:
            raise ValueError("theta_max cannot be negative")
        if v > THETA_MAX_DEFAULT:
            raise ValueError(
                f"theta_max cannot be greater than {multiples_of_pi(THETA_MAX_DEFAULT)}"
            )
        return v

    @validator("theta_min")
    def check_theta_min(cls, v):
        if v < THETA_MIN_DEFAULT:
            raise ValueError("theta_min cannot be negative")
        if v > THETA_MAX_DEFAULT:
            raise ValueError(
                f"theta_min cannot be greater than {multiples_of_pi(THETA_MAX_DEFAULT)}"
            )
        return v



import numpy as np
import pyssht as ssht
from numpy import typing as npt
from pys2let import axisym_wav_l

from sleplet.slepian.slepian_functions import SlepianFunctions
from sleplet.utils.convolution_methods import sifting_convolution
from sleplet.utils.slepian_methods import compute_s_p_omega


def slepian_wavelet_forward(
    f_p: npt.NDArray[np.complex_ | np.float_],
    wavelets: npt.NDArray[np.float_],
    shannon: int,
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    computes the coefficients of the given tiling function in Slepian space
    """
    return find_non_zero_wavelet_coefficients(
        sifting_convolution(wavelets, f_p, shannon=shannon), axis=1
    )


def slepian_wavelet_inverse(
    wav_coeffs: npt.NDArray[np.complex_ | np.float_],
    wavelets: npt.NDArray[np.float_],
    shannon: int,
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    computes the inverse wavelet transform in Slepian space
    """
    # ensure wavelets are the same shape as the coefficients
    wavelets_shannon = wavelets[: len(wav_coeffs)]
    wavelet_reconstruction = sifting_convolution(
        wavelets_shannon, wav_coeffs.T, shannon=shannon
    )
    return wavelet_reconstruction.sum(axis=0)


def axisymmetric_wavelet_forward(
    L: int,
    flm: npt.NDArray[np.complex_ | np.float_],
    wavelets: npt.NDArray[np.complex_],
) -> npt.NDArray[np.complex_]:
    """
    computes the coefficients of the axisymmetric wavelets
    """
    w = np.zeros(wavelets.shape, dtype=np.complex_)
    for ell in range(L):
        ind_m0 = ssht.elm2ind(ell, 0)
        wav_0 = np.sqrt((4 * np.pi) / (2 * ell + 1)) * wavelets[:, ind_m0].conj()
        for m in range(-ell, ell + 1):
            ind = ssht.elm2ind(ell, m)
            w[:, ind] = wav_0 * flm[ind]
    return w


def axisymmetric_wavelet_inverse(
    L: int, wav_coeffs: npt.NDArray[np.complex_], wavelets: npt.NDArray[np.complex_]
) -> npt.NDArray[np.complex_]:
    """
    computes the inverse axisymmetric wavelet transform
    """
    flm = np.zeros(L**2, dtype=np.complex_)
    for ell in range(L):
        ind_m0 = ssht.elm2ind(ell, 0)
        wav_0 = np.sqrt((4 * np.pi) / (2 * ell + 1)) * wavelets[:, ind_m0]
        for m in range(-ell, ell + 1):
            ind = ssht.elm2ind(ell, m)
            flm[ind] = (wav_coeffs[:, ind] * wav_0).sum()
    return flm


def compute_wavelet_covariance(
    L: int, wavelets: npt.NDArray[np.complex_], *, var_signal: float
) -> npt.NDArray[np.float_]:
    """
    computes the theoretical covariance of the wavelet coefficients
    """
    covar_theory = (np.abs(wavelets) ** 2).sum(axis=1)
    return covar_theory * var_signal


def compute_slepian_wavelet_covariance(
    L: int,
    wavelets: npt.NDArray[np.float_],
    slepian: SlepianFunctions,
    *,
    var_signal: float,
) -> npt.NDArray[np.float_]:
    """
    computes the theoretical covariance of the wavelet coefficients
    """
    s_p = compute_s_p_omega(L, slepian)
    wavelets_reshape = wavelets[:, : slepian.N, np.newaxis, np.newaxis]
    covar_theory = (np.abs(wavelets_reshape) ** 2 * np.abs(s_p) ** 2).sum(axis=1)
    return covar_theory * var_signal


def create_axisymmetric_wavelets(
    L: int, B: int, j_min: int
) -> npt.NDArray[np.complex_]:
    """
    computes the axisymmetric wavelets
    """
    kappas = create_kappas(L, B, j_min)
    wavelets = np.zeros((kappas.shape[0], L**2), dtype=np.complex_)
    for ell in range(L):
        factor = np.sqrt((2 * ell + 1) / (4 * np.pi))
        ind = ssht.elm2ind(ell, 0)
        wavelets[:, ind] = factor * kappas[:, ell]
    return wavelets


def create_kappas(xlim: int, B: int, j_min: int) -> npt.NDArray[np.float_]:
    """
    computes the Slepian wavelets
    """
    kappa0, kappa = axisym_wav_l(B, xlim, j_min)
    return np.concatenate((kappa0[np.newaxis], kappa.T))


def find_non_zero_wavelet_coefficients(
    wav_coeffs: npt.NDArray[np.complex_ | np.float_], *, axis: int | tuple[int, ...]
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    finds the coefficients within the shannon number to speed up computations
    """
    return wav_coeffs[wav_coeffs.any(axis=axis)]



import numpy as np
import pyssht as ssht
from hypothesis import given, seed, settings
from hypothesis.strategies import SearchStrategy, floats
from numpy.testing import assert_allclose, assert_equal, assert_raises

from sleplet.functions.flm.dirac_delta import DiracDelta
from sleplet.utils.plot_methods import calc_nearest_grid_point
from sleplet.utils.slepian_methods import slepian_inverse
from sleplet.utils.vars import RANDOM_SEED, SAMPLING_SCHEME

L = 128
THETA_MAX = np.pi / 3


def valid_alphas() -> SearchStrategy[float]:
    """
    alpha can be in the range [0, 2*pi)
    """
    return floats(min_value=0, max_value=2, exclude_max=True)


def valid_betas() -> SearchStrategy[float]:
    """
    beta can be in the range [0, pi]
    """
    return floats(min_value=0, max_value=1)


@seed(RANDOM_SEED)
@settings(max_examples=8, deadline=None)
@given(alpha_pi_frac=valid_alphas(), beta_pi_frac=valid_betas())
def test_dirac_delta_rotate_translate(alpha_pi_frac, beta_pi_frac) -> None:
    """
    test to ensure that rotation and translation
    give the same result for the Dirac delta
    """
    dd = DiracDelta(L)
    alpha, beta = calc_nearest_grid_point(L, alpha_pi_frac, beta_pi_frac)
    dd_rot = dd.rotate(alpha, beta)
    dd_trans = dd.translate(alpha, beta)
    assert_allclose(np.abs(dd_trans - dd_rot).mean(), 0, atol=0)


def test_slepian_translation_changes_max_polar(slepian_dirac_delta_polar_cap) -> None:
    """
    test to ensure the location of the maximum of a field moves when translated
    """
    _, beta = calc_nearest_grid_point(
        slepian_dirac_delta_polar_cap.L, 0, THETA_MAX / np.pi
    )
    sdd_trans = slepian_dirac_delta_polar_cap.translate(
        slepian_dirac_delta_polar_cap.alpha,
        beta,
        shannon=slepian_dirac_delta_polar_cap.slepian.N,
    )
    field = slepian_inverse(
        sdd_trans,
        slepian_dirac_delta_polar_cap.L,
        slepian_dirac_delta_polar_cap.slepian,
    )
    new_max = tuple(np.argwhere(field == field.max())[0])
    thetas, _ = ssht.sample_positions(
        slepian_dirac_delta_polar_cap.L, Grid=True, Method=SAMPLING_SCHEME
    )
    assert_raises(
        AssertionError,
        assert_equal,
        slepian_dirac_delta_polar_cap.beta,
        thetas[new_max],
    )


def test_slepian_translation_changes_max_lim_lat_lon(
    slepian_dirac_delta_lim_lat_lon,
) -> None:
    """
    test to ensure the location of the maximum of a field moves when translated
    """
    _, beta = calc_nearest_grid_point(
        slepian_dirac_delta_lim_lat_lon.L, 0, THETA_MAX / np.pi
    )
    sdd_trans = slepian_dirac_delta_lim_lat_lon.translate(
        slepian_dirac_delta_lim_lat_lon.alpha,
        beta,
        shannon=slepian_dirac_delta_lim_lat_lon.slepian.N,
    )
    field = slepian_inverse(
        sdd_trans,
        slepian_dirac_delta_lim_lat_lon.L,
        slepian_dirac_delta_lim_lat_lon.slepian,
    )
    new_max = tuple(np.argwhere(field == field.max())[0])
    thetas, _ = ssht.sample_positions(
        slepian_dirac_delta_lim_lat_lon.L, Grid=True, Method=SAMPLING_SCHEME
    )
    assert_raises(
        AssertionError,
        assert_equal,
        slepian_dirac_delta_lim_lat_lon.beta,
        thetas[new_max],
    )



from abc import abstractmethod
from dataclasses import KW_ONLY

import numpy as np
from numpy import typing as npt
from pydantic import validator
from pydantic.dataclasses import dataclass

from sleplet.meshes.classes.mesh import Mesh
from sleplet.utils.mask_methods import ensure_masked_bandlimit_mesh_signal
from sleplet.utils.string_methods import filename_args
from sleplet.utils.validation import Validation

COEFFICIENTS_TO_NOT_MASK: str = "slepian"


@dataclass(config=Validation)
class MeshCoefficients:
    mesh: Mesh
    _: KW_ONLY
    extra_args: list[int] | None = None
    noise: float | None = None
    region: bool = False

    def __post_init_post_parse__(self) -> None:
        self._setup_args()
        self.name = self._create_name()
        self.coefficients = self._create_coefficients()
        self._add_details_to_name()
        self.unnoised_coefficients, self.snr = self._add_noise_to_signal()

    def _add_details_to_name(self) -> None:
        """
        adds region to the name if present if not a Slepian function
        """
        if self.region and "slepian" not in self.mesh.name:
            self.name += "_region"
        if self.noise is not None:
            self.name += f"{filename_args(self.noise, 'noise')}"
        if self.mesh.zoom:
            self.name += "_zoom"

    @validator("coefficients", check_fields=False)
    def check_coefficients(cls, v, values):
        if (
            values["region"]
            and COEFFICIENTS_TO_NOT_MASK not in cls.__class__.__name__.lower()
        ):
            v = ensure_masked_bandlimit_mesh_signal(values["mesh"], v)
        return v

    @abstractmethod
    def _add_noise_to_signal(
        self,
    ) -> tuple[npt.NDArray[np.complex_ | np.float_] | None, float | None]:
        """
        adds Gaussian white noise to the signal
        """
        raise NotImplementedError

    @abstractmethod
    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        """
        creates the flm on the north pole
        """
        raise NotImplementedError

    @abstractmethod
    def _create_name(self) -> str:
        """
        creates the name of the function
        """
        raise NotImplementedError

    @abstractmethod
    def _setup_args(self) -> None:
        """
        initialises function specific args
        either default value or user input
        """
        raise NotImplementedError



import numpy as np
import pyssht as ssht
from numpy.testing import assert_allclose, assert_array_equal, assert_equal
from pys2let import pys2let_j_max

from sleplet.functions.flm.axisymmetric_wavelet_coefficients_earth import (
    AxisymmetricWaveletCoefficientsEarth,
)
from sleplet.functions.flm.axisymmetric_wavelet_coefficients_south_america import (
    AxisymmetricWaveletCoefficientsSouthAmerica,
)
from sleplet.utils.slepian_methods import slepian_forward
from sleplet.utils.vars import SAMPLING_SCHEME
from sleplet.utils.wavelet_methods import (
    axisymmetric_wavelet_inverse,
    compute_slepian_wavelet_covariance,
    compute_wavelet_covariance,
    create_kappas,
    find_non_zero_wavelet_coefficients,
    slepian_wavelet_forward,
    slepian_wavelet_inverse,
)

B = 2
J_MIN = 0
L_LARGE = 128
L_SMALL = 16
VAR_SIGNAL = 1


def test_synthesis_polar(slepian_wavelets_polar_cap, earth_polar_cap) -> None:
    """
    tests that Slepian polar wavelet synthesis matches the coefficients
    """
    coefficients = slepian_forward(
        slepian_wavelets_polar_cap.L,
        slepian_wavelets_polar_cap.slepian,
        flm=earth_polar_cap.coefficients,
    )
    wav_coeffs = slepian_wavelet_forward(
        coefficients,
        slepian_wavelets_polar_cap.wavelets,
        slepian_wavelets_polar_cap.slepian.N,
    )
    f_p = slepian_wavelet_inverse(
        wav_coeffs,
        slepian_wavelets_polar_cap.wavelets,
        slepian_wavelets_polar_cap.slepian.N,
    )
    assert_allclose(np.abs(f_p - coefficients).mean(), 0, atol=1e-14)


def test_synthesis_lim_lat_lon(slepian_wavelets_lim_lat_lon, earth_lim_lat_lon) -> None:
    """
    tests that Slepian lim_lat_lon wavelet synthesis matches the coefficients
    """
    coefficients = slepian_forward(
        slepian_wavelets_lim_lat_lon.L,
        slepian_wavelets_lim_lat_lon.slepian,
        flm=earth_lim_lat_lon.coefficients,
    )
    wav_coeffs = slepian_wavelet_forward(
        coefficients,
        slepian_wavelets_lim_lat_lon.wavelets,
        slepian_wavelets_lim_lat_lon.slepian.N,
    )
    f_p = slepian_wavelet_inverse(
        wav_coeffs,
        slepian_wavelets_lim_lat_lon.wavelets,
        slepian_wavelets_lim_lat_lon.slepian.N,
    )
    assert_allclose(np.abs(f_p - coefficients).mean(), 0, atol=0)


def test_axisymmetric_synthesis_earth() -> None:
    """
    tests that the axisymmetric wavelet synthesis recoveres the coefficients
    """
    awc = AxisymmetricWaveletCoefficientsEarth(L_SMALL, B=B, j_min=J_MIN)
    flm = axisymmetric_wavelet_inverse(L_SMALL, awc.wavelet_coefficients, awc.wavelets)
    assert_allclose(np.abs(flm - awc.earth.coefficients).mean(), 0, atol=1e-13)


def test_axisymmetric_synthesis_south_america() -> None:
    """
    tests that the axisymmetric wavelet synthesis recoveres the coefficients
    """
    awc = AxisymmetricWaveletCoefficientsSouthAmerica(L_SMALL, B=B, j_min=J_MIN)
    flm = axisymmetric_wavelet_inverse(L_SMALL, awc.wavelet_coefficients, awc.wavelets)
    assert_allclose(np.abs(flm - awc.south_america.coefficients).mean(), 0, atol=1e-14)


def test_only_wavelet_coefficients_within_shannon_returned() -> None:
    """
    verifies that only the non-zero wavelet coefficients are returned
    """
    coeffs_in = np.array([[3], [2], [1], [0]])
    coeffs_out = np.array([[3], [2], [1]])
    shannon_coeffs = find_non_zero_wavelet_coefficients(coeffs_in, axis=1)
    assert_array_equal(shannon_coeffs, coeffs_out)


def test_create_kappas() -> None:
    """
    checks that the method creates the scaling function and wavelets
    """
    wavelets = create_kappas(L_LARGE**2, B, J_MIN)
    j_max = pys2let_j_max(B, L_LARGE**2, J_MIN)
    assert_equal(j_max - J_MIN + 2, wavelets.shape[0])
    assert_equal(L_LARGE**2, wavelets.shape[1])


def test_wavelet_covariance(random_nd_flm) -> None:
    """
    checks that sigma^j is computed for the axisymmetric case
    """
    covariance = compute_wavelet_covariance(
        L_SMALL, random_nd_flm, var_signal=VAR_SIGNAL
    )
    assert_equal(random_nd_flm.shape[0], covariance.shape[0])


def test_slepian_wavelet_covariance(slepian_wavelets_polar_cap) -> None:
    """
    checks that sigma^j is computed for the Slepian case
    """
    covariance = compute_slepian_wavelet_covariance(
        slepian_wavelets_polar_cap.L,
        slepian_wavelets_polar_cap.wavelets,
        slepian_wavelets_polar_cap.slepian,
        var_signal=VAR_SIGNAL,
    )
    assert_equal(slepian_wavelets_polar_cap.wavelets.shape[0], covariance.shape[0])
    assert_equal(
        ssht.sample_shape(slepian_wavelets_polar_cap.L, Method=SAMPLING_SCHEME),
        covariance.shape[1:],
    )



from dataclasses import KW_ONLY, field
from pathlib import Path

import cmocean
import numpy as np
import plotly.offline as py
import pyssht as ssht
from numpy import typing as npt
from plotly.graph_objs import Figure, Surface
from plotly.graph_objs.surface import Lighting
from pydantic.dataclasses import dataclass

from sleplet import logger
from sleplet.utils.plot_methods import (
    boost_field,
    calc_plot_resolution,
    convert_colourscale,
    create_plot_type,
    normalise_function,
    set_outside_region_to_minimum,
)
from sleplet.utils.plotly_methods import (
    create_camera,
    create_colour_bar,
    create_layout,
    create_tick_mark,
)
from sleplet.utils.region import Region
from sleplet.utils.validation import Validation
from sleplet.utils.vars import SAMPLING_SCHEME, SPHERE_UNSEEN

MW_POLE_LENGTH = 2

_fig_path = Path(__file__).resolve().parents[1] / "figures"


@dataclass(config=Validation)
class Plot:
    f: npt.NDArray[np.complex_ | np.float_]
    L: int
    filename: str
    _: KW_ONLY
    amplitude: float | None = None
    annotations: list[dict] = field(default_factory=list)
    normalise: bool = True
    plot_type: str = "real"
    reality: bool = False
    region: Region | None = None
    spin: int = 0
    upsample: bool = True

    def __post_init_post_parse__(self) -> None:
        self.resolution = calc_plot_resolution(self.L) if self.upsample else self.L
        if self.upsample:
            self.filename += f"_res{self.resolution}"
        self.filename += f"_{self.plot_type}"
        if self.normalise:
            self.filename += "_norm"

    def execute(self) -> None:
        """
        creates basic plotly plot rather than matplotlib
        """
        f = self._prepare_field(self.f)

        # get values from the setup
        x, y, z, f_plot, vmin, vmax = self._setup_plot(
            f, self.resolution, method=SAMPLING_SCHEME
        )

        if isinstance(self.region, Region):
            # make plot area clearer
            f_plot = set_outside_region_to_minimum(f_plot, self.resolution, self.region)

        # appropriate zoom in on north pole
        camera = create_camera(-0.1, -0.1, 10, 7.88)

        # pick largest tick max value
        tick_mark = create_tick_mark(vmin, vmax, amplitude=self.amplitude)

        data = [
            Surface(
                x=x,
                y=y,
                z=z,
                surfacecolor=f_plot,
                cmax=1 if self.normalise else tick_mark,
                cmid=0.5 if self.normalise else 0,
                cmin=0 if self.normalise else -tick_mark,
                colorbar=create_colour_bar(tick_mark, normalise=self.normalise),
                colorscale=convert_colourscale(cmocean.cm.ice),
                lighting=Lighting(ambient=1),
                reversescale=True,
            )
        ]

        layout = create_layout(camera, annotations=self.annotations)

        fig = Figure(data=data, layout=layout)

        html_filename = str(_fig_path / "html" / f"{self.filename}.html")

        py.plot(fig, filename=html_filename)

        for file_type in {"png", "pdf"}:
            filename = str(_fig_path / file_type / f"{self.filename}.{file_type}")
            logger.info(f"saving {filename}")
            fig.write_image(filename, engine="kaleido")

    @staticmethod
    def _setup_plot(
        f: npt.NDArray[np.float_],
        resolution: int,
        *,
        method: str = "MW",
        close: bool = True,
        parametric: bool = False,
        parametric_scaling: list[float] | None = None,
        color_range: list[float] | None = None,
    ) -> tuple[
        npt.NDArray[np.float_],
        npt.NDArray[np.float_],
        npt.NDArray[np.float_],
        npt.NDArray[np.float_],
        float,
        float,
    ]:
        """
        function which creates the data for the matplotlib/plotly plot
        """
        if parametric_scaling is None:
            parametric_scaling = [0.0, 0.5]
        if method == "MW_pole":
            if len(f) == MW_POLE_LENGTH:
                f, _ = f
            else:
                f, _, _ = f

        thetas, phis = ssht.sample_positions(resolution, Grid=True, Method=method)

        if thetas.size != f.size:
            raise AttributeError("Bandlimit L deos not match that of f")

        f_plot = f.copy()

        f_max = f_plot.max()
        f_min = f_plot.min()

        if not isinstance(color_range, list):
            vmin = f_min
            vmax = f_max
        else:
            vmin = color_range[0]
            vmax = color_range[1]
            f_plot[f_plot < color_range[0]] = color_range[0]
            f_plot[f_plot > color_range[1]] = color_range[1]
            f_plot[f_plot == SPHERE_UNSEEN] = np.nan

        # Compute position scaling for parametric plot.
        f_normalised = (
            (f_plot - vmin / (vmax - vmin)) * parametric_scaling[1]
            + parametric_scaling[0]
            if parametric
            else np.zeros(f_plot.shape)
        )

        # Close plot.
        if close:
            _, n_phi = ssht.sample_shape(resolution, Method=method)
            f_plot = np.insert(f_plot, n_phi, f[:, 0], axis=1)
            if parametric:
                f_normalised = np.insert(
                    f_normalised, n_phi, f_normalised[:, 0], axis=1
                )
            thetas = np.insert(thetas, n_phi, thetas[:, 0], axis=1)
            phis = np.insert(phis, n_phi, phis[:, 0], axis=1)

        # Compute location of vertices.
        x, y, z = (
            ssht.spherical_to_cart(f_normalised, thetas, phis)
            if parametric
            else ssht.s2_to_cart(thetas, phis)
        )

        return x, y, z, f_plot, vmin, vmax

    def _prepare_field(
        self, f: npt.NDArray[np.complex_ | np.float_]
    ) -> npt.NDArray[np.float_]:
        """
        boosts, forces plot type and then scales the field before plotting
        """
        boosted_field = boost_field(
            f,
            self.L,
            self.resolution,
            reality=self.reality,
            spin=self.spin,
            upsample=self.upsample,
        )
        field_space = create_plot_type(boosted_field, self.plot_type)
        return normalise_function(field_space, normalise=self.normalise)



from collections.abc import Callable
from typing import Any

import numpy as np
import pyssht as ssht
from numpy import typing as npt
from numpy.random import Generator

from sleplet.meshes.classes.mesh import Mesh
from sleplet.utils.integration_methods import integrate_whole_mesh
from sleplet.utils.vars import (
    AFRICA_ALPHA,
    AFRICA_BETA,
    AFRICA_GAMMA,
    SAMPLING_SCHEME,
    SOUTH_AMERICA_ALPHA,
    SOUTH_AMERICA_BETA,
    SOUTH_AMERICA_GAMMA,
)


def create_spherical_harmonic(L: int, ind: int) -> npt.NDArray[np.complex_]:
    """
    create a spherical harmonic in harmonic space for the given index
    """
    flm = np.zeros(L**2, dtype=np.complex_)
    flm[ind] = 1
    return flm


def boost_coefficient_resolution(flm: npt.NDArray[Any], boost: int) -> npt.NDArray[Any]:
    """
    calculates a boost in resolution for given flm
    """
    return np.pad(flm, (0, boost), "constant")


def invert_flm_boosted(
    flm: npt.NDArray[np.complex_],
    L: int,
    resolution: int,
    *,
    reality: bool = False,
    spin: int = 0,
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    performs the inverse harmonic transform
    """
    boost = resolution**2 - L**2
    flm = boost_coefficient_resolution(flm, boost)
    return ssht.inverse(
        flm, resolution, Reality=reality, Spin=spin, Method=SAMPLING_SCHEME
    )


def ensure_f_bandlimited(
    grid_fun: Callable[
        [npt.NDArray[np.float_], npt.NDArray[np.float_]], npt.NDArray[np.float_]
    ],
    L: int,
    *,
    reality: bool,
    spin: int,
) -> npt.NDArray[np.complex_]:
    """
    if the function created is created in pixel space rather than harmonic
    space then need to transform it into harmonic space first before using it
    """
    thetas, phis = ssht.sample_positions(L, Grid=True, Method=SAMPLING_SCHEME)
    f = grid_fun(thetas, phis)
    return ssht.forward(f, L, Reality=reality, Spin=spin, Method=SAMPLING_SCHEME)


def create_emm_vector(L: int) -> npt.NDArray[np.float_]:
    """
    create vector of m values for a given L
    """
    emm = np.zeros(2 * L * 2 * L)
    k = 0

    for ell in range(2 * L):
        M = 2 * ell + 1
        emm[k : k + M] = np.arange(-ell, ell + 1)
        k += M
    return emm


def compute_random_signal(
    L: int, rng: Generator, *, var_signal: float
) -> npt.NDArray[np.complex_]:
    """
    generates a normally distributed random signal of a
    complex signal with mean 0 and variance 1
    """
    return np.sqrt(var_signal / 2) * (
        rng.standard_normal(L**2) + 1j * rng.standard_normal(L**2)
    )


def mesh_forward(
    mesh: Mesh, u: npt.NDArray[np.complex_ | np.float_]
) -> npt.NDArray[np.float_]:
    """
    computes the mesh forward transform from real space to harmonic space
    """
    u_i = np.zeros(mesh.mesh_eigenvalues.shape[0])
    for i, phi_i in enumerate(mesh.basis_functions):
        u_i[i] = integrate_whole_mesh(mesh.vertices, mesh.faces, u, phi_i)
    return u_i


def mesh_inverse(
    mesh: Mesh, u_i: npt.NDArray[np.complex_ | np.float_]
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    computes the mesh inverse transform from harmonic space to real space
    """
    return (u_i[:, np.newaxis] * mesh.basis_functions).sum(axis=0)


def rotate_earth_to_south_america(
    earth_flm: npt.NDArray[np.complex_ | np.float_], L: int
) -> npt.NDArray[np.complex_]:
    """
    rotates the flms of the Earth to a view centered on South America
    """
    return ssht.rotate_flms(
        earth_flm, SOUTH_AMERICA_ALPHA, SOUTH_AMERICA_BETA, SOUTH_AMERICA_GAMMA, L
    )


def rotate_earth_to_africa(
    earth_flm: npt.NDArray[np.complex_ | np.float_], L: int
) -> npt.NDArray[np.complex_]:
    """
    rotates the flms of the Earth to a view centered on Africa
    """
    return ssht.rotate_flms(earth_flm, AFRICA_ALPHA, AFRICA_BETA, AFRICA_GAMMA, L)



from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from numpy import typing as npt

from sleplet import logger
from sleplet.slepian.slepian_region.slepian_polar_cap import SlepianPolarCap
from sleplet.utils.plot_methods import save_plot

L = 16
LEGEND_POS = (0, 0)
RANKS = 60
TEXT_BOX: dict[str, str | float] = {"boxstyle": "round", "color": "w"}
THETA_RANGE = {10: (0, 0), 20: (0, 1), 30: (1, 0), 40: (1, 1)}

fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")


def main() -> None:
    """
    creates a plot of Slepian eigenvalues against rank
    """
    x = len(THETA_RANGE) // 2
    _, ax = plt.subplots(x, x, sharex="col", sharey="row")
    for theta_max, position in THETA_RANGE.items():
        _create_plot(ax, position, theta_max)
    ax[LEGEND_POS].legend(ncol=2)
    save_plot(fig_path, "simons_5_3")


def _create_plot(ax: npt.NDArray, position: tuple[int, int], theta_max: int) -> None:
    """
    helper method which actually makes the plot
    """
    logger.info(f"theta_max={theta_max}")
    slepian = SlepianPolarCap(L, np.deg2rad(theta_max))
    axs = ax[position]
    legend = "full" if position == LEGEND_POS else False
    assert isinstance(slepian.order, np.ndarray)  # noqa: S101
    orders = np.abs(slepian.order[:RANKS])
    labels = np.array([f"$\\pm${m}" if m != 0 else m for m in orders])
    idx = np.argsort(orders)
    sns.scatterplot(
        x=range(RANKS),
        y=slepian.eigenvalues[:RANKS],
        hue=labels,
        hue_order=labels[idx],
        legend=legend,
        style=labels,
        ax=axs,
    )
    axs.axvline(x=slepian.N - 1)
    axs.annotate(
        f"N={slepian.N}",
        xy=(slepian.N - 1, 1),
        xytext=(0, 7),
        ha="center",
        textcoords="offset points",
        annotation_clip=False,
    )
    if position[1] == 0:
        axs.set_ylabel("$\\mu$")
    if position[0] == 1:
        axs.set_xlabel("$p$")
    axs.text(
        0.02,
        0.34,
        rf"$\Theta={{{theta_max}}}^\circ$",
        transform=axs.transAxes,
        bbox=TEXT_BOX,
    )


if __name__ == "__main__":
    main()



import numpy as np
import pyssht as ssht
from numpy import typing as npt
from pydantic import validator
from pydantic.dataclasses import dataclass

from sleplet.functions.f_lm import F_LM
from sleplet.utils.harmonic_methods import create_spherical_harmonic
from sleplet.utils.string_methods import convert_camel_case_to_snake_case, filename_args
from sleplet.utils.validation import Validation


@dataclass(config=Validation, kw_only=True)
class SphericalHarmonic(F_LM):
    ell: int = 0
    m: int = 0

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        ind = ssht.elm2ind(self.ell, self.m)
        return create_spherical_harmonic(self.L, ind)

    def _create_name(self) -> str:
        return (
            f"{convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{filename_args(self.ell, 'l')}"
            f"{filename_args(self.m, 'm')}"
        )

    def _set_reality(self) -> bool:
        return False

    def _set_spin(self) -> int:
        return 0

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 2
            if len(self.extra_args) != num_args:
                raise ValueError(f"The number of extra arguments should be {num_args}")
            self.ell, self.m = self.extra_args

    @validator("ell")
    def check_ell(cls, v, values):
        if not isinstance(v, int):
            raise TypeError("ell should be an integer")
        if v < 0:
            raise ValueError("ell should be positive")
        if v >= values["L"]:
            raise ValueError("ell should be less than or equal to L")
        return v

    @validator("m")
    def check_m(cls, v, values):
        if not isinstance(v, int):
            raise TypeError("m should be an integer")
        if abs(v) > values["ell"]:
            raise ValueError("the magnitude of m should be less than ell")
        return v



"""
.. include:: ../../README.md
"""

import logging

from sleplet._version import __version__

logging.basicConfig(
    format="[%(asctime)s] [%(levelname)s] --- %(message)s (%(filename)s:%(lineno)s)",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO,
)
logger = logging.getLogger(__name__)



#!/usr/bin/env python
from argparse import ArgumentParser, Namespace

import numpy as np
import pyssht as ssht
from numpy import typing as npt

from sleplet import logger
from sleplet.functions.coefficients import Coefficients
from sleplet.plotting.create_plot_sphere import Plot
from sleplet.utils.class_lists import COEFFICIENTS, MAPS_LM
from sleplet.utils.harmonic_methods import (
    rotate_earth_to_africa,
    rotate_earth_to_south_america,
)
from sleplet.utils.mask_methods import create_default_region
from sleplet.utils.plot_methods import calc_nearest_grid_point
from sleplet.utils.slepian_methods import slepian_forward, slepian_inverse
from sleplet.utils.string_methods import (
    convert_classes_list_to_snake_case,
    filename_angle,
)
from sleplet.utils.vars import (
    ALPHA_DEFAULT,
    ANNOTATION_COLOUR,
    ARROW_STYLE,
    BETA_DEFAULT,
    SAMPLING_SCHEME,
)


def valid_maps(map_name: str) -> str:
    """
    check if valid map
    """
    if map_name in convert_classes_list_to_snake_case(MAPS_LM):
        return map_name
    else:
        raise ValueError(f"{map_name} is not a valid map to convolve")


def valid_plotting(func_name: str) -> str:
    """
    check if valid function
    """
    if func_name in convert_classes_list_to_snake_case(COEFFICIENTS):
        return func_name
    else:
        raise ValueError(f"{func_name} is not a valid function to plot")


def read_args() -> Namespace:
    """
    method to read args from the command line
    """
    parser = ArgumentParser(description="Create SSHT plot")
    parser.add_argument(
        "function",
        type=valid_plotting,
        choices=convert_classes_list_to_snake_case(COEFFICIENTS),
        help="function to plot on the sphere",
    )
    parser.add_argument(
        "--alpha",
        "-a",
        type=float,
        default=ALPHA_DEFAULT,
        help=f"alpha/phi pi fraction - defaults to {ALPHA_DEFAULT}",
    )
    parser.add_argument("--bandlimit", "-L", type=int, default=16, help="bandlimit")
    parser.add_argument(
        "--beta",
        "-b",
        type=float,
        default=BETA_DEFAULT,
        help=f"beta/theta pi fraction - defaults to {BETA_DEFAULT}",
    )
    parser.add_argument(
        "--convolve",
        "-c",
        type=valid_maps,
        default=None,
        choices=convert_classes_list_to_snake_case(MAPS_LM),
        help="glm to perform sifting convolution with i.e. flm x glm*",
    )
    parser.add_argument(
        "--extra_args",
        "-e",
        type=int,
        nargs="+",
        help="list of extra args for functions",
    )
    parser.add_argument(
        "--gamma",
        "-g",
        type=float,
        default=0,
        help="gamma pi fraction - defaults to 0 - rotation only",
    )
    parser.add_argument(
        "--method",
        "-m",
        type=str,
        nargs="?",
        default="north",
        const="north",
        choices=["north", "rotate", "translate"],
        help="plotting routine: defaults to north",
    )
    parser.add_argument("--noise", "-n", type=int, help="the SNR_IN of the noise level")
    parser.add_argument(
        "--outline",
        "-o",
        action="store_false",
        help="flag which removes any annotation",
    )
    parser.add_argument(
        "--region",
        "-r",
        action="store_true",
        help="flag which masks the function for a region",
    )
    parser.add_argument(
        "--smoothing",
        "-s",
        type=int,
        help="the scaling of the sigma in Gaussian smoothing of the Earth",
    )
    parser.add_argument(
        "--type",
        "-t",
        type=str,
        nargs="?",
        default="real",
        const="real",
        choices=["abs", "real", "imag", "sum"],
        help="plotting type: defaults to real",
    )
    parser.add_argument(
        "--unnormalise",
        "-u",
        action="store_true",
        help="flag turns off normalisation for plot",
    )
    parser.add_argument(
        "--unzeropad",
        "-z",
        action="store_true",
        help="flag turns off upsampling for plot",
    )
    parser.add_argument(
        "--view",
        "-v",
        type=str,
        nargs="?",
        default="south_america",
        const="south_america",
        choices=["africa", "south_america"],
        help="view of Earth: defaults to 'south_america'",
    )
    return parser.parse_args()


def plot(
    f: Coefficients,
    g: Coefficients | None,
    *,
    alpha_pi_frac: float,
    beta_pi_frac: float,
    gamma_pi_frac: float,
    annotations: bool,
    normalise: bool,
    method: str,
    plot_type: str,
    upsample: bool,
    earth_view: str,
    amplitude: float | None,
) -> None:
    """
    master plotting method
    """
    filename = f.name
    coefficients = f.coefficients

    # turn off annotation if needed
    logger.info(f"annotations on: {annotations}")
    annotation = []

    # Shannon number for Slepian coefficients
    shannon = f.slepian.N if hasattr(f, "slepian") else None

    logger.info(f"plotting method: '{method}'")
    match method:
        case "rotate":
            coefficients, filename = _rotation_helper(
                f, filename, alpha_pi_frac, beta_pi_frac, gamma_pi_frac
            )
        case "translate":
            coefficients, filename, trans_annotation = _translation_helper(
                f, filename, alpha_pi_frac, beta_pi_frac, shannon
            )

            # annotate translation point
            if annotations:
                annotation.append(trans_annotation)

    if g is not None:
        coefficients, filename = _convolution_helper(
            f, g, coefficients, shannon, filename
        )

    # rotate plot of Earth
    if "earth" in filename:
        match earth_view:
            case "africa":
                coefficients = rotate_earth_to_africa(coefficients, f.L)
                filename += "_africa"
            case "south_america":
                coefficients = rotate_earth_to_south_america(coefficients, f.L)

    # get field value
    field = _coefficients_to_field(f, coefficients)

    # do plot
    Plot(
        field,
        f.L,
        filename,
        amplitude=amplitude,
        annotations=annotation,
        normalise=normalise,
        plot_type=plot_type,
        reality=f.reality,
        region=f.region if hasattr(f, "slepian") else None,
        spin=f.spin,
        upsample=upsample,
    ).execute()


def _rotation_helper(
    f: Coefficients,
    filename: str,
    alpha_pi_frac: float,
    beta_pi_frac: float,
    gamma_pi_frac: float,
) -> tuple[npt.NDArray[np.complex_], str]:
    """
    performs the rotation specific steps
    """
    logger.info(
        "angles: (alpha, beta, gamma) = "
        f"({alpha_pi_frac}, {beta_pi_frac}, {gamma_pi_frac})"
    )
    filename += f"_rotate_{filename_angle(alpha_pi_frac, beta_pi_frac, gamma_pi_frac)}"

    # calculate angles
    alpha, beta = calc_nearest_grid_point(f.L, alpha_pi_frac, beta_pi_frac)
    gamma = gamma_pi_frac * np.pi

    # rotate by alpha, beta, gamma
    coefficients = f.rotate(alpha, beta, gamma=gamma)
    return coefficients, filename


def _translation_helper(
    f: Coefficients,
    filename: str,
    alpha_pi_frac: float,
    beta_pi_frac: float,
    shannon: int | None,
) -> tuple[npt.NDArray[np.complex_ | np.float_], str, dict]:
    """
    performs the translation specific steps
    """
    logger.info(f"angles: (alpha, beta) = ({alpha_pi_frac}, {beta_pi_frac})")
    # don't add gamma if translation
    filename += f"_translate_{filename_angle(alpha_pi_frac, beta_pi_frac)}"

    # calculate angles
    alpha, beta = calc_nearest_grid_point(f.L, alpha_pi_frac, beta_pi_frac)

    # translate by alpha, beta
    coefficients = f.translate(alpha, beta, shannon=shannon)

    # annotate translation point
    x, y, z = ssht.s2_to_cart(beta, alpha)
    annotation = {
        **{"x": x, "y": y, "z": z, "arrowcolor": ANNOTATION_COLOUR},
        **ARROW_STYLE,
    }
    return coefficients, filename, annotation


def _convolution_helper(
    f: Coefficients,
    g: Coefficients,
    coefficients: npt.NDArray[np.complex_ | np.float_],
    shannon: int | None,
    filename: str,
) -> tuple[npt.NDArray[np.complex_ | np.float_], str]:
    """
    performs the convolution specific steps
    """
    g_coefficients = (
        slepian_forward(f.L, f.slepian, flm=g.coefficients)
        if hasattr(f, "slepian")
        else g.coefficients
    )
    coefficients = f.convolve(g_coefficients, coefficients, shannon=shannon)

    filename += f"_convolved_{g.name}"
    return coefficients, filename


def _coefficients_to_field(
    f: Coefficients, coefficients: npt.NDArray[np.complex_ | np.float_]
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    computes the field over the samples from the harmonic/Slepian coefficients
    """
    return (
        slepian_inverse(coefficients, f.L, f.slepian)
        if hasattr(f, "slepian")
        else ssht.inverse(
            coefficients, f.L, Reality=f.reality, Spin=f.spin, Method=SAMPLING_SCHEME
        )
    )


def compute_amplitude_for_noisy_plots(f: Coefficients) -> float | None:
    """
    for the noised plots fix the amplitude to the initial data
    """
    return (
        np.abs(_coefficients_to_field(f, f.unnoised_coefficients)).max()
        if f.unnoised_coefficients is not None
        else None
    )


def main() -> None:
    args = read_args()

    mask = create_default_region() if args.region else None

    f = COEFFICIENTS[
        convert_classes_list_to_snake_case(COEFFICIENTS).index(args.function)
    ](
        args.bandlimit,
        extra_args=args.extra_args,
        region=mask,
        noise=args.noise if args.noise is not None else None,
        smoothing=args.smoothing if args.smoothing is not None else None,
    )

    g = (
        MAPS_LM[convert_classes_list_to_snake_case(MAPS_LM).index(args.convolve)](
            args.bandlimit
        )
        if isinstance(args.convolve, str)
        else None
    )

    # custom amplitude for noisy plots
    amplitude = compute_amplitude_for_noisy_plots(f)

    plot(
        f,
        g,
        alpha_pi_frac=args.alpha,
        beta_pi_frac=args.beta,
        gamma_pi_frac=args.gamma,
        annotations=args.outline,
        normalise=not args.unnormalise,
        method=args.method,
        plot_type=args.type,
        upsample=not args.unzeropad,
        earth_view=args.view,
        amplitude=amplitude,
    )


if __name__ == "__main__":
    main()



import pyssht as ssht
from numpy.testing import assert_equal

from sleplet.utils.harmonic_methods import (
    boost_coefficient_resolution,
    invert_flm_boosted,
)
from sleplet.utils.vars import SAMPLING_SCHEME

L_LARGE = 128
L_SMALL = 16


def test_harmonic_coefficients_padded(random_flm) -> None:
    """
    tests that harmonic coefficients are zero padded for plotting
    """
    boost = L_LARGE**2 - L_SMALL**2
    flm_boosted = boost_coefficient_resolution(random_flm, boost)
    assert_equal(len(flm_boosted), L_LARGE**2)


def test_invert_flm_and_boost(random_flm) -> None:
    """
    tests that the flm has been boosted and has right shape
    """
    n_theta, n_phi = ssht.sample_shape(L_LARGE, Method=SAMPLING_SCHEME)
    f = invert_flm_boosted(random_flm, L_SMALL, L_LARGE)
    assert_equal(f.shape, (n_theta, n_phi))



from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from numpy import typing as npt

from sleplet.utils.plot_methods import save_plot

fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")

DELTA_T = 0.001
FREQUENCIES = [5, 10, 15]
LENGTH = 0.512


def main() -> None:
    """
    plots a Dirac impulse and it's Fourier transform
    """
    size = int(LENGTH / DELTA_T)
    amplitude = np.zeros(size * len(FREQUENCIES))
    for c, f in enumerate(FREQUENCIES):
        amplitude[size * c : size * (c + 1)] = _ricker(f)
    plt.plot(amplitude)
    plt.xticks([])
    plt.xlabel(r"$t$")
    save_plot(fig_path, "ricker_wavelets")


def _ricker(freq: float) -> npt.NDArray[np.float_]:
    """
    creates a Ricker wavelet
    """
    t = np.arange(-LENGTH / 2, (LENGTH - DELTA_T) / 2, DELTA_T)
    return (1.0 - 2.0 * (np.pi**2) * (freq**2) * (t**2)) * np.exp(
        -(np.pi**2) * (freq**2) * (t**2)
    )


if __name__ == "__main__":
    main()



from denoising_slepian_mesh import main

from sleplet import logger

MESH_SNR_DICT = {
    "cheetah": -8.64,
    "dragon": -8.12,
    "bird": -5.17,
    "teapot": -3.11,
    "cube": -7.33,
    "homer": -5,
}
SIGMA = 2

if __name__ == "__main__":
    for mesh, snr in MESH_SNR_DICT.items():
        logger.info(f"\n\n\ndenoising {mesh}")
        main(mesh, snr, SIGMA)



from argparse import ArgumentParser
from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt

from sleplet.meshes.classes.mesh import Mesh
from sleplet.meshes.classes.mesh_slepian import MeshSlepian
from sleplet.scripts.plotting_on_mesh import valid_meshes
from sleplet.utils.class_lists import MESHES
from sleplet.utils.plot_methods import save_plot

fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")


def main(mesh_name: str, num_basis_fun: int) -> None:
    """
    plots the Slepian eigenvalues of the given mesh
    """
    mesh = Mesh(
        mesh_name,
        number_basis_functions=num_basis_fun,
    )
    mesh_slepian = MeshSlepian(mesh)
    plt.semilogx(mesh_slepian.slepian_eigenvalues, "k.")
    plt.axvline(mesh_slepian.N, c="k", ls="--", alpha=0.8)
    plt.annotate(
        f"N={mesh_slepian.N}",
        xy=(mesh_slepian.N, 1),
        xytext=(17, 3),
        ha="center",
        textcoords="offset points",
        annotation_clip=False,
    )
    ticks = 2 ** np.arange(np.log2(mesh.number_basis_functions) + 1, dtype=int)
    plt.xticks(ticks, ticks)
    plt.xlabel(r"$p$")
    plt.ylabel(r"$\mu$")
    save_plot(
        fig_path, f"{mesh_name}_slepian_eigenvalues_b{mesh.number_basis_functions}"
    )


if __name__ == "__main__":
    parser = ArgumentParser(description="mesh Slepian eigenvalues")
    parser.add_argument(
        "function",
        type=valid_meshes,
        choices=MESHES,
        help="mesh to plot",
    )
    parser.add_argument(
        "--basis",
        "-b",
        type=int,
        help="number of basis functions to extract",
    )
    args = parser.parse_args()
    main(args.function, args.basis)



from sleplet.functions.fp.slepian_africa import SlepianAfrica
from sleplet.plotting.create_plot_sphere import Plot
from sleplet.utils.region import Region
from sleplet.utils.slepian_methods import choose_slepian_method, slepian_inverse
from sleplet.utils.vars import SMOOTHING

L = 128
NORMALISE = False


def main() -> None:
    """
    the reconstruction of a signal in Slepian space
    """
    region = Region(mask_name="africa")
    slepian = choose_slepian_method(L, region)
    africa = SlepianAfrica(L, region=region, smoothing=SMOOTHING)

    # perform reconstruction
    f = slepian_inverse(africa.coefficients, L, slepian)

    # plot
    name = f"africa_slepian_reconstruction_L{L}"
    Plot(f, L, name, normalise=NORMALISE, region=slepian.region).execute()


if __name__ == "__main__":
    main()



from argparse import ArgumentParser

import numpy as np

from sleplet import logger
from sleplet.meshes.classes.mesh import Mesh
from sleplet.meshes.slepian_coefficients.mesh_slepian_field import MeshSlepianField
from sleplet.meshes.slepian_coefficients.mesh_slepian_wavelets import (
    MeshSlepianWavelets,
)
from sleplet.plotting.create_plot_mesh import Plot
from sleplet.scripts.plotting_on_mesh import valid_meshes
from sleplet.utils.class_lists import MESHES
from sleplet.utils.denoising import denoising_mesh_slepian
from sleplet.utils.slepian_methods import slepian_mesh_inverse
from sleplet.utils.string_methods import filename_args

B = 3
J_MIN = 2
N_SIGMA = 1
NORMALISE = False
SNR_IN = -5


def main(mesh_name: str, snr: float, sigma: int) -> None:
    """
    denoising demo using Slepian wavelets
    """
    logger.info(f"SNR={snr}, n_sigma={sigma}")
    # setup
    mesh = Mesh(mesh_name, zoom=True)

    # create map & noised map
    fun = MeshSlepianField(mesh)
    fun_noised = MeshSlepianField(mesh, noise=snr)

    # create wavelets
    smw = MeshSlepianWavelets(mesh, B=B, j_min=J_MIN)

    # fix amplitude
    amplitude = np.abs(
        slepian_mesh_inverse(fun_noised.mesh_slepian, fun.coefficients)
    ).max()

    f = denoising_mesh_slepian(fun, fun_noised, smw, snr, sigma)
    name = f"{mesh_name}{filename_args(snr, 'snr')}{filename_args(sigma,'n')}_denoised"
    Plot(mesh, name, f, amplitude=amplitude, normalise=NORMALISE, region=True).execute()


if __name__ == "__main__":
    parser = ArgumentParser(description="denoising")
    parser.add_argument(
        "function",
        type=valid_meshes,
        choices=MESHES,
        help="mesh to plot",
    )
    parser.add_argument(
        "--noise",
        "-n",
        type=int,
        default=SNR_IN,
    )
    parser.add_argument(
        "--sigma",
        "-s",
        type=int,
        default=N_SIGMA,
    )
    args = parser.parse_args()
    main(args.function, args.noise, args.sigma)



from pathlib import Path

import numpy as np
import pyssht as ssht
from matplotlib import colors
from matplotlib import pyplot as plt
from numpy import typing as npt

from sleplet import logger
from sleplet.functions.coefficients import Coefficients
from sleplet.utils.harmonic_methods import invert_flm_boosted
from sleplet.utils.mask_methods import create_mask_region
from sleplet.utils.region import Region
from sleplet.utils.slepian_methods import slepian_inverse
from sleplet.utils.vars import (
    SAMPLING_SCHEME,
    SPHERE_UNSEEN,
)


def calc_plot_resolution(L: int) -> int:
    """
    calculate appropriate resolution for given L
    """
    res_dict = {1: 6, 2: 5, 3: 4, 7: 3, 9: 2, 10: 1}

    return next(
        (
            L * 2**exponent
            for log_bandlimit, exponent in res_dict.items()
            if 2**log_bandlimit > L
        ),
        L,
    )


def convert_colourscale(
    cmap: colors, *, pl_entries: int = 255
) -> list[tuple[float, str]]:
    """
    converts cmocean colourscale to a plotly colourscale
    """
    h = 1 / (pl_entries - 1)
    pl_colorscale = []

    for k in range(pl_entries):
        C = list(map(np.uint8, np.array(cmap(k * h)[:3]) * 255))
        pl_colorscale.append((k * h, f"rgb{(C[0], C[1], C[2])}"))

    return pl_colorscale


def calc_nearest_grid_point(
    L: int, alpha_pi_fraction: float, beta_pi_fraction: float
) -> tuple[float, float]:
    """
    calculate nearest index of alpha/beta for translation
    this is due to calculating omega' through the pixel
    values - the translation needs to be at the same position
    as the rotation such that the difference error is small
    """
    thetas, phis = ssht.sample_positions(L, Method=SAMPLING_SCHEME)
    pix_j = np.abs(phis - alpha_pi_fraction * np.pi).argmin()
    pix_i = np.abs(thetas - beta_pi_fraction * np.pi).argmin()
    alpha, beta = phis[pix_j], thetas[pix_i]
    logger.info(f"grid point: (alpha, beta)=({alpha:e}, {beta:e})")
    return alpha, beta


def save_plot(path: Path, name: str) -> None:
    """
    helper method to save plots
    """
    plt.tight_layout()
    for file_type in {"png", "pdf"}:
        filename = path / file_type / f"{name}.{file_type}"
        logger.info(f"saving {filename}")
        plt.savefig(filename, bbox_inches="tight")
    plt.show()


def find_max_amplitude(
    function: Coefficients, *, plot_type: str = "real", upsample: bool = True
) -> float:
    """
    for a given set of coefficients it finds the largest absolute value for a
    given plot type such that plots can have the same scale as the input
    """
    # compute inverse transform
    if hasattr(function, "slepian"):
        field = slepian_inverse(function.coefficients, function.L, function.slepian)
    else:
        field = ssht.inverse(function.coefficients, function.L, Method=SAMPLING_SCHEME)

    # find resolution of final plot for boosting if necessary
    resolution = calc_plot_resolution(function.L) if upsample else function.L

    # boost field to match final plot
    boosted_field = boost_field(
        field,
        function.L,
        resolution,
        reality=function.reality,
        spin=function.spin,
        upsample=upsample,
    )

    # find maximum absolute value for given plot type
    return np.abs(create_plot_type(boosted_field, plot_type)).max()


def create_plot_type(
    field: npt.NDArray[np.complex_ | np.float_], plot_type: str
) -> npt.NDArray[np.float_]:
    """
    gets the given plot type of the field
    """
    logger.info(f"plotting type: '{plot_type}'")
    plot_dict = {
        "abs": np.abs(field),
        "imag": field.imag,
        "real": field.real,
        "sum": field.real + field.imag,
    }
    return plot_dict[plot_type]


def set_outside_region_to_minimum(
    f_plot: npt.NDArray[np.float_], L: int, region: Region
) -> npt.NDArray[np.float_]:
    """
    for the Slepian region set the outisde area to negative infinity
    hence it is clear we are only interested in the coloured region
    """
    # create mask of interest
    mask = create_mask_region(L, region)

    # adapt for closed plot
    _, n_phi = ssht.sample_shape(L, Method=SAMPLING_SCHEME)
    closed_mask = np.insert(mask, n_phi, mask[:, 0], axis=1)

    # set values outside mask to negative infinity
    return np.where(closed_mask, f_plot, SPHERE_UNSEEN)


def normalise_function(
    f: npt.NDArray[np.float_], *, normalise: bool
) -> npt.NDArray[np.float_]:
    """
    normalise function between 0 and 1 for visualisation
    """
    if not normalise:
        return f
    elif (f == 0).all():
        # if all 0, set to 0
        return f + 0.5
    elif np.allclose(f, f.max()):
        # if all non-zero, set to 1
        return f / f.max()
    else:
        # scale from [0, 1]
        return (f - f.min()) / f.ptp()


def boost_field(
    field: npt.NDArray[np.complex_ | np.float_],
    L: int,
    resolution: int,
    *,
    reality: bool,
    spin: int,
    upsample: bool,
) -> npt.NDArray[np.complex_ | np.float_]:
    """
    inverts and then boosts the field before plotting
    """
    if not upsample:
        return field
    flm = ssht.forward(field, L, Reality=reality, Spin=spin, Method=SAMPLING_SCHEME)
    return invert_flm_boosted(flm, L, resolution, reality=reality, spin=spin)



from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from numpy import typing as npt

from sleplet import logger
from sleplet.data.setup_pooch import find_on_pooch_then_local
from sleplet.slepian.slepian_region.slepian_polar_cap import SlepianPolarCap
from sleplet.utils.harmonic_methods import invert_flm_boosted
from sleplet.utils.plot_methods import calc_plot_resolution, save_plot
from sleplet.utils.vars import THETA_MAX_DEFAULT, THETA_MIN_DEFAULT

COLUMNS = 3
L = 16
ORDER = 0
PHI_IDX = 0
ROWS = 2
RESOLUTION = calc_plot_resolution(L)
SIGNS = [1, -1, 1, -1, 1, -1]
TEXT_BOX: dict[str, str | float] = {"boxstyle": "round", "color": "w"}
THETA_MAX = 40


fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")


def main() -> None:
    """
    create fig 5.1 from Spatiospectral Concentration on a Sphere by Simons et al 2006
    """
    x = np.linspace(THETA_MIN_DEFAULT, np.rad2deg(THETA_MAX_DEFAULT), RESOLUTION + 1)
    i = (x < THETA_MAX).sum()
    _, ax = plt.subplots(ROWS, COLUMNS, sharex="col", sharey="row")
    axes = ax.flatten()
    plt.setp(
        ax,
        xlim=[0, 180],
        xticks=[0, 40, 180],
        xticklabels=[r"$0^\circ$", r"$40^\circ$", r"$180^\circ$"],
        ylim=[-3, 3],
        yticks=[-2, 0, 2],
    )
    slepian = SlepianPolarCap(L, np.deg2rad(THETA_MAX), order=ORDER)
    for rank in range(ROWS * COLUMNS):
        _helper(axes[rank], slepian, x, i, rank)
    save_plot(fig_path, "slepian_colatitude")


def _helper(
    ax: plt.Axes,
    slepian: SlepianPolarCap,
    x: npt.NDArray[np.float_],
    i: int,
    rank: int,
) -> None:
    """
    helper which plots the required order and specified ranks
    """
    logger.info(f"plotting rank={rank}")
    flm = slepian.eigenvectors[rank] * SIGNS[rank]
    lam = slepian.eigenvalues[rank]
    f = invert_flm_boosted(flm, L, RESOLUTION).real
    if rank > COLUMNS - 1:
        ax.set_xlabel(r"colatitude $\theta$")
    ax.plot(x[:i], f[:i, PHI_IDX], x[i:], f[i:, PHI_IDX])
    p = _find_p_value(rank, slepian.N)
    ax.text(
        0.39,
        0.92,
        rf"$\mu_{{{p+1}}}={{{lam:.6f}}}$",
        transform=ax.transAxes,
        bbox=TEXT_BOX,
    )


def _find_p_value(rank: int, shannon: int) -> int:
    """
    method to find the effective p rank of the Slepian function
    """
    orders = np.load(
        find_on_pooch_then_local(
            f"slepian_eigensolutions_D_polar{THETA_MAX}_L{L}_N{shannon}_orders.npy"
        )
    )
    return np.where(orders == ORDER)[0][rank]


if __name__ == "__main__":
    main()



import numpy as np
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet.functions.f_lm import F_LM
from sleplet.utils.harmonic_methods import ensure_f_bandlimited
from sleplet.utils.string_methods import convert_camel_case_to_snake_case, filename_args
from sleplet.utils.validation import Validation
from sleplet.utils.vars import PHI_0, THETA_0


@dataclass(config=Validation, kw_only=True)
class ElongatedGaussian(F_LM):
    p_sigma: float = 0.1
    t_sigma: float = 1

    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        return ensure_f_bandlimited(
            self._grid_fun, self.L, reality=self.reality, spin=self.spin
        )

    def _create_name(self) -> str:
        return (
            f"{convert_camel_case_to_snake_case(self.__class__.__name__)}"
            f"{filename_args(self.t_sigma, 'tsig')}"
            f"{filename_args(self.p_sigma, 'psig')}"
        )

    def _set_reality(self) -> bool:
        return True

    def _set_spin(self) -> int:
        return 0

    def _setup_args(self) -> None:
        if isinstance(self.extra_args, list):
            num_args = 2
            if len(self.extra_args) != num_args:
                raise ValueError(f"The number of extra arguments should be {num_args}")
            self.t_sigma, self.p_sigma = (
                np.float_power(10, x) for x in self.extra_args
            )

    def _grid_fun(
        self, theta: npt.NDArray[np.float_], phi: npt.NDArray[np.float_]
    ) -> npt.NDArray[np.float_]:
        """
        function on the grid
        """
        return np.exp(
            -(
                ((theta - THETA_0) / self.t_sigma) ** 2
                + ((phi - PHI_0) / self.p_sigma) ** 2
            )
            / 2
        )



from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt

from sleplet.slepian.slepian_region.slepian_arbitrary import SlepianArbitrary
from sleplet.utils.plot_methods import save_plot

fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")

COLOURS = ["b", "k"]
L = 128
REGIONS = ["africa", "south_america"]
X_POS = [19, -16]
ZORDER = [2, 1]


def main() -> None:
    """
    plots the tiling of the Slepian line
    """
    for c, region in enumerate(REGIONS):
        slepian = SlepianArbitrary(L, region)
        p_range = np.arange(0, L**2)
        plt.semilogx(
            p_range,
            slepian.eigenvalues,
            f"{COLOURS[c]}.",
            label=" ".join(region.title().split("_")),
            zorder=ZORDER[c],
        )
        plt.axvline(slepian.N, color=COLOURS[c], ls="--", alpha=0.8)
        plt.annotate(
            f"N={slepian.N}",
            xy=(slepian.N, 1),
            xytext=(X_POS[c], 3),
            ha="center",
            textcoords="offset points",
            annotation_clip=False,
            color=COLOURS[c],
        )
    ticks = 2 ** np.arange(np.log2(L**2) + 1, dtype=int)
    plt.xticks(ticks, ticks)
    plt.xlabel(r"$p$")
    plt.ylabel(r"$\mu$")
    plt.legend(loc=3)
    save_plot(fig_path, f"combined_eigenvalues_L{L}")


if __name__ == "__main__":
    main()



from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from scipy.interpolate import pchip

from sleplet.slepian.slepian_region.slepian_arbitrary import SlepianArbitrary
from sleplet.utils.plot_methods import save_plot
from sleplet.utils.wavelet_methods import create_kappas

fig_path = Path(__file__).resolve().parents[3] / "src" / "sleplet" / "figures"
sns.set(context="paper")

B = 3
J_MIN = 2
L = 128
STEP = 0.01


def main() -> None:
    """
    plots the tiling of the Slepian line
    """
    xlim = L**2
    x = np.arange(xlim)
    xi = np.arange(0, xlim - 1 + STEP, STEP)
    kappas = create_kappas(xlim, B, J_MIN)
    yi = pchip(x, kappas[0])
    plt.semilogx(xi, yi(xi), label=r"$\Phi_p$")
    for j, k in enumerate(kappas[1:]):
        yi = pchip(x, k)
        plt.semilogx(xi, yi(xi), label=rf"$\Psi^{{{j+J_MIN}}}_p$")
    slepian = SlepianArbitrary(L, "africa")
    plt.axvline(slepian.N, color="k", linestyle="dashed")
    plt.annotate(
        f"N={slepian.N}",
        xy=(slepian.N, 1),
        xytext=(19, 3),
        ha="center",
        textcoords="offset points",
        annotation_clip=False,
    )
    plt.xlim(1, xlim)
    ticks = 2 ** np.arange(np.log2(xlim) + 1, dtype=int)
    plt.xticks(ticks, ticks)
    plt.xlabel(r"$p$")
    plt.legend()
    save_plot(fig_path, f"africa_slepian_tiling_L{L}")


if __name__ == "__main__":
    main()



from pathlib import Path

import numpy as np
import pyssht as ssht
import seaborn as sns
from matplotlib import pyplot as plt

from sleplet import logger
from sleplet.functions.flm.harmonic_gaussian import HarmonicGaussian
from sleplet.utils.plot_methods import save_plot
from sleplet.utils.vars import ALPHA_DEFAULT, SAMPLING_SCHEME

fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")

L = 128


def compute_translation_normalisation_theta() -> None:
    """
    analysis of the translation norm for referee
    """
    hg = HarmonicGaussian(L)
    thetas, _ = ssht.sample_positions(L, Method=SAMPLING_SCHEME)
    norm = np.zeros(len(thetas))
    for i, theta in enumerate(thetas):
        logger.info(f"compute norm {i+1}/{len(thetas)}")
        ylm_omega_prime = ssht.create_ylm(theta, ALPHA_DEFAULT, L).reshape(L**2)
        norm[i] = np.sqrt((np.abs(hg.coefficients * ylm_omega_prime) ** 2).sum())
    plt.plot(np.rad2deg(thetas), norm)
    plt.xlabel(r"$\theta/^\circ$")
    plt.ylabel(r"${\Vert f_{\ell m} Y_{\ell m}(\omega')\Vert}_{2}$")
    plt.xlim(0, 180)
    plt.ylim(0, np.ceil(norm.max()))
    save_plot(fig_path, f"harmonic_gaussian_translation_normalisation_L{L}")


if __name__ == "__main__":
    compute_translation_normalisation_theta()



from dataclasses import KW_ONLY
from pathlib import Path

import cmocean
import numpy as np
import plotly.offline as py
from matplotlib.colors import LinearSegmentedColormap
from numpy import typing as npt
from plotly.graph_objs import Figure, Mesh3d
from plotly.graph_objs.mesh3d import Lighting
from pydantic.dataclasses import dataclass

from sleplet import logger
from sleplet.meshes.classes.mesh import Mesh
from sleplet.utils.mask_methods import convert_region_on_vertices_to_faces
from sleplet.utils.mesh_methods import average_functions_on_vertices_to_faces
from sleplet.utils.plot_methods import convert_colourscale, normalise_function
from sleplet.utils.plotly_methods import (
    create_colour_bar,
    create_layout,
    create_tick_mark,
)
from sleplet.utils.validation import Validation
from sleplet.utils.vars import (
    MESH_CBAR_FONT_SIZE,
    MESH_CBAR_LEN,
    MESH_UNSEEN,
)

_fig_path = Path(__file__).resolve().parents[1] / "figures"


@dataclass(config=Validation)
class Plot:
    mesh: Mesh
    filename: str
    f: npt.NDArray[np.complex_ | np.float_]
    _: KW_ONLY
    amplitude: float | None = None
    colour: LinearSegmentedColormap = cmocean.cm.ice
    normalise: bool = True
    region: bool = False

    def __post_init_post_parse__(self) -> None:
        if self.normalise:
            self.filename += "_norm"

    def execute(self) -> None:
        """
        creates 3d plotly mesh plot
        """
        vmin, vmax = self.f.min(), self.f.max()
        f = self._prepare_field(self.f)

        if self.region:
            # make plot area clearer
            f = self._set_outside_region_to_minimum(f)

        # pick largest tick max value
        tick_mark = create_tick_mark(vmin, vmax, amplitude=self.amplitude)

        data = [
            Mesh3d(
                x=self.mesh.vertices[:, 0],
                y=self.mesh.vertices[:, 2],
                z=self.mesh.vertices[:, 1],
                i=self.mesh.faces[:, 0],
                j=self.mesh.faces[:, 1],
                k=self.mesh.faces[:, 2],
                intensitymode="cell",
                intensity=f,
                cmax=1 if self.normalise else tick_mark,
                cmid=0.5 if self.normalise else 0,
                cmin=0 if self.normalise else -tick_mark,
                colorbar=create_colour_bar(
                    tick_mark,
                    normalise=self.normalise,
                    bar_len=MESH_CBAR_LEN,
                    bar_pos=self.mesh.colourbar_pos,
                    font_size=MESH_CBAR_FONT_SIZE,
                ),
                colorscale=convert_colourscale(self.colour),
                lighting=Lighting(ambient=1),
                reversescale=True,
            )
        ]

        layout = create_layout(self.mesh.camera_view)

        fig = Figure(data=data, layout=layout)

        html_filename = str(_fig_path / "html" / f"{self.filename}.html")

        py.plot(fig, filename=html_filename)

        for file_type in {"png", "pdf"}:
            filename = str(_fig_path / file_type / f"{self.filename}.{file_type}")
            logger.info(f"saving {filename}")
            fig.write_image(filename, engine="kaleido")

    def _prepare_field(
        self, f: npt.NDArray[np.complex_ | np.float_]
    ) -> npt.NDArray[np.float_]:
        """
        scales the field before plotting
        """
        return normalise_function(
            average_functions_on_vertices_to_faces(self.mesh.faces, f),
            normalise=self.normalise,
        )

    def _set_outside_region_to_minimum(
        self, f: npt.NDArray[np.float_]
    ) -> npt.NDArray[np.float_]:
        """
        for the Slepian region set the outisde area to negative infinity
        hence it is clear we are only interested in the coloured region
        """
        region_on_faces = convert_region_on_vertices_to_faces(self.mesh)
        return np.where(region_on_faces, f, MESH_UNSEEN)



from abc import abstractmethod

import numpy as np
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet.functions.coefficients import Coefficients
from sleplet.utils.mask_methods import create_default_region
from sleplet.utils.noise import compute_snr, create_slepian_noise
from sleplet.utils.region import Region
from sleplet.utils.slepian_methods import choose_slepian_method, compute_s_p_omega_prime
from sleplet.utils.validation import Validation


@dataclass(config=Validation)
class F_P(Coefficients):  # noqa: N801
    def __post_init_post_parse__(self) -> None:
        self.region: Region | None = (
            self.region if isinstance(self.region, Region) else create_default_region()
        )
        self.slepian = choose_slepian_method(self.L, self.region)
        super().__post_init_post_parse__()

    def rotate(
        self, alpha: float, beta: float, *, gamma: float = 0
    ) -> npt.NDArray[np.complex_]:
        raise NotImplementedError("Slepian rotation is not defined")

    def _translation_helper(
        self, alpha: float, beta: float
    ) -> npt.NDArray[np.complex_]:
        return compute_s_p_omega_prime(self.L, alpha, beta, self.slepian).conj()

    def _add_noise_to_signal(
        self,
    ) -> tuple[npt.NDArray[np.complex_ | np.float_] | None, float | None]:
        """
        adds Gaussian white noise converted to Slepian space
        """
        self.coefficients: npt.NDArray[np.complex_ | np.float_]
        if self.noise is not None:
            unnoised_coefficients = self.coefficients.copy()
            n_p = create_slepian_noise(
                self.L, self.coefficients, self.slepian, self.noise
            )
            snr = compute_snr(self.coefficients, n_p, "Slepian")
            self.coefficients = self.coefficients + n_p
            return unnoised_coefficients, snr
        return None, None

    @abstractmethod
    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        raise NotImplementedError

    @abstractmethod
    def _create_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def _set_reality(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def _set_spin(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def _setup_args(self) -> None:
        raise NotImplementedError



import numpy as np
from numpy import typing as npt

from sleplet.meshes.classes.mesh import Mesh

MACHINE_EPSILON = 1e-14


def clean_evals_and_evecs(
    eigendecomposition: tuple,
) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
    """
    need eigenvalues and eigenvectors to be in a certain format
    """
    # access values
    eigenvalues, eigenvectors = eigendecomposition

    # eigenvalues should be real
    eigenvalues = eigenvalues.real

    # Sort eigenvalues and eigenvectors in descending order of eigenvalues
    idx = eigenvalues.argsort()[::-1]
    eigenvalues = eigenvalues[idx]
    eigenvectors = eigenvectors[:, idx].conj().T

    # ensure first element of each eigenvector is positive
    eigenvectors *= np.where(eigenvectors[:, 0] < 0, -1, 1)[:, np.newaxis]

    # find repeating eigenvalues and ensure orthorgonality
    pairs = np.where(np.abs(np.diff(eigenvalues)) < MACHINE_EPSILON)[0] + 1
    eigenvectors[pairs] *= 1j

    return eigenvalues, eigenvectors


def compute_mesh_shannon(mesh: Mesh) -> int:
    """
    computes the effective Shannon number for a region of a mesh
    """
    num_basis_fun = mesh.mesh_eigenvalues.shape[0]
    region_vertices = mesh.region.sum()
    total_vertices = mesh.region.shape[0]
    return round(region_vertices / total_vertices * num_basis_fun)



import numpy as np

from sleplet import logger
from sleplet.functions.fp.slepian import Slepian
from sleplet.plotting.create_plot_sphere import Plot
from sleplet.utils.region import Region
from sleplet.utils.slepian_methods import slepian_inverse

L = 16
NORMALISE = False
RANKS = 16
THETA_MAX = 40


def main() -> None:
    """
    create fig 5.4 from Spatiospectral Concentration on a Sphere by Simons et al 2006
    """
    region = Region(theta_max=np.deg2rad(THETA_MAX))
    for r in range(RANKS):
        logger.info(f"plotting rank={r}")
        slepian_function = Slepian(L, rank=r, region=region)
        f = slepian_inverse(slepian_function.coefficients, L, slepian_function.slepian)
        Plot(
            f,
            L,
            slepian_function.name,
            normalise=NORMALISE,
            region=slepian_function.region,
        ).execute()


if __name__ == "__main__":
    main()



from abc import abstractmethod

import numpy as np
import pyssht as ssht
from numpy import typing as npt
from pydantic.dataclasses import dataclass

from sleplet.functions.coefficients import Coefficients
from sleplet.utils.noise import compute_snr, create_noise
from sleplet.utils.validation import Validation


@dataclass(config=Validation)
class F_LM(Coefficients):  # noqa: N801
    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def rotate(
        self, alpha: float, beta: float, *, gamma: float = 0
    ) -> npt.NDArray[np.complex_]:
        return ssht.rotate_flms(self.coefficients, alpha, beta, gamma, self.L)

    def _translation_helper(
        self, alpha: float, beta: float
    ) -> npt.NDArray[np.complex_]:
        return ssht.create_ylm(beta, alpha, self.L).conj().flatten()

    def _add_noise_to_signal(
        self,
    ) -> tuple[npt.NDArray[np.complex_ | np.float_] | None, float | None]:
        """
        adds Gaussian white noise to the signal
        """
        self.coefficients: npt.NDArray[np.complex_ | np.float_]
        if self.noise is not None:
            unnoised_coefficients = self.coefficients.copy()
            nlm = create_noise(self.L, self.coefficients, self.noise)
            snr = compute_snr(self.coefficients, nlm, "Harmonic")
            self.coefficients = self.coefficients + nlm
            return unnoised_coefficients, snr
        return None, None

    @abstractmethod
    def _create_coefficients(self) -> npt.NDArray[np.complex_ | np.float_]:
        raise NotImplementedError

    @abstractmethod
    def _create_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def _set_reality(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def _set_spin(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def _setup_args(self) -> None:
        raise NotImplementedError



from argparse import ArgumentParser
from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from scipy.interpolate import pchip

from sleplet.meshes.classes.mesh import Mesh
from sleplet.meshes.classes.mesh_slepian import MeshSlepian
from sleplet.scripts.plotting_on_mesh import valid_meshes
from sleplet.utils.class_lists import MESHES
from sleplet.utils.plot_methods import save_plot
from sleplet.utils.wavelet_methods import create_kappas

fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"
sns.set(context="paper")

B = 3
J_MIN = 2
STEP = 0.01


def main(mesh_name: str) -> None:
    """
    plots the tiling of the Slepian line
    """
    # initialise mesh and Slepian mesh
    mesh = Mesh(mesh_name)
    mesh_slepian = MeshSlepian(mesh)

    # set up x-axis
    xlim = mesh.mesh_eigenvalues.shape[0]
    x = np.arange(xlim)

    # scaling function
    xi = np.arange(0, xlim - 1 + STEP, STEP)
    kappas = create_kappas(xlim, B, J_MIN)
    yi = pchip(x, kappas[0])
    plt.semilogx(xi, yi(xi), label=r"$\Phi_p$")

    # wavelets
    for j, k in enumerate(kappas[1:]):
        yi = pchip(x, k)
        plt.semilogx(xi, yi(xi), label=rf"$\Psi^{{{j+J_MIN}}}_p$")

    # add vertical line
    plt.axvline(mesh_slepian.N, color="k", linestyle="dashed")
    plt.annotate(
        f"N={mesh_slepian.N}",
        xy=(mesh_slepian.N, 1),
        xytext=(17, 3),
        ha="center",
        textcoords="offset points",
        annotation_clip=False,
    )

    # format plot
    plt.xlim(1, xlim)
    ticks = 2 ** np.arange(np.log2(xlim) + 1, dtype=int)
    plt.xticks(ticks, ticks)
    plt.xlabel(r"$p$")
    plt.legend()
    save_plot(fig_path, f"{mesh_name}_slepian_tiling_b{mesh.mesh_eigenvalues.shape[0]}")


if __name__ == "__main__":
    parser = ArgumentParser(description="mesh tiling")
    parser.add_argument(
        "function",
        type=valid_meshes,
        choices=MESHES,
        help="mesh to plot",
    )
    args = parser.parse_args()
    main(args.function)



from concurrent.futures import ThreadPoolExecutor
from dataclasses import KW_ONLY
from pathlib import Path

import gmpy2 as gp
import numpy as np
import pyssht as ssht
from numpy import linalg as LA  # noqa: N812
from numpy import typing as npt
from pydantic import validator
from pydantic.dataclasses import dataclass

from sleplet import logger
from sleplet.data.setup_pooch import find_on_pooch_then_local
from sleplet.slepian.slepian_functions import SlepianFunctions
from sleplet.utils.harmonic_methods import create_emm_vector
from sleplet.utils.mask_methods import create_mask_region
from sleplet.utils.parallel_methods import (
    attach_to_shared_memory_block,
    create_shared_memory_array,
    free_shared_memory,
    release_shared_memory,
    split_arr_into_chunks,
)
from sleplet.utils.region import Region
from sleplet.utils.validation import Validation
from sleplet.utils.vars import NCPU

L_SAVE_ALL = 16

_data_path = Path(__file__).resolve().parents[2] / "data"


@dataclass(config=Validation)
class SlepianPolarCap(SlepianFunctions):
    theta_max: float
    _: KW_ONLY
    gap: bool = False
    order: int | npt.NDArray[np.int_] | None = None

    def __post_init_post_parse__(self) -> None:
        super().__post_init_post_parse__()

    def _create_fn_name(self) -> str:
        return f"slepian_{self.region.name_ending}"

    def _create_region(self) -> Region:
        return Region(gap=self.gap, theta_max=self.theta_max)

    def _create_mask(self) -> npt.NDArray[np.float_]:
        return create_mask_region(self.L, self.region)

    def _calculate_area(self) -> float:
        return 2 * np.pi * (1 - np.cos(self.theta_max))

    def _create_matrix_location(self) -> str:
        return f"slepian_eigensolutions_D_{self.region.name_ending}_L{self.L}_N{self.N}"

    def _solve_eigenproblem(
        self,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        eval_loc = f"{self.matrix_location}_eigenvalues.npy"
        evec_loc = f"{self.matrix_location}_eigenvectors.npy"
        order_loc = f"{self.matrix_location}_orders.npy"
        try:
            return self._solve_eigenproblem_from_files(eval_loc, evec_loc, order_loc)
        except TypeError:
            return self._solve_eigenproblem_from_scratch(eval_loc, evec_loc, order_loc)

    def _solve_eigenproblem_from_files(
        self, eval_loc: str, evec_loc: str, order_loc: str
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        """
        solves eigenproblem with files already saved
        """
        eigenvalues = np.load(find_on_pooch_then_local(eval_loc))
        eigenvectors = np.load(find_on_pooch_then_local(evec_loc))
        orders = np.load(find_on_pooch_then_local(order_loc))

        if self.order is not None:
            idx = np.where(orders == self.order)
            return eigenvalues[idx], eigenvectors[idx]
        else:
            self.order = orders
            return eigenvalues, eigenvectors

    def _solve_eigenproblem_from_scratch(
        self, eval_loc: str, evec_loc: str, order_loc: str
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        """
        sovles eigenproblem from scratch and then saves the files
        """
        if isinstance(self.order, int):
            return self._solve_eigenproblem_order(self.order)

        evals_all = np.empty(0)
        evecs_all = np.empty((0, self.L**2), dtype=np.complex_)
        emm = np.empty(0, dtype=int)
        for m in range(-(self.L - 1), self.L):
            evals_m, evecs_m = self._solve_eigenproblem_order(m)
            evals_all = np.append(evals_all, evals_m)
            evecs_all = np.concatenate((evecs_all, evecs_m))
            emm = np.append(emm, [m] * len(evals_m))
        (
            eigenvalues,
            eigenvectors,
            self.order,
        ) = self._sort_all_evals_and_evecs(evals_all, evecs_all, emm)
        limit = self.N if self.L > L_SAVE_ALL else None
        np.save(_data_path / eval_loc, eigenvalues)
        np.save(_data_path / evec_loc, eigenvectors[:limit])
        np.save(_data_path / order_loc, self.order)
        return eigenvalues, eigenvectors

    def _solve_eigenproblem_order(
        self, m: int
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        """
        solves the eigenproblem for a given order 'm;
        """
        emm = create_emm_vector(self.L)
        Dm = self._create_Dm_matrix(abs(m), emm)
        eigenvalues, gl = LA.eigh(Dm)
        eigenvalues, eigenvectors = self._clean_evals_and_evecs(eigenvalues, gl, emm, m)
        return eigenvalues, eigenvectors

    def _sort_all_evals_and_evecs(
        self,
        eigenvalues: npt.NDArray[np.float_],
        eigenvectors: npt.NDArray[np.complex_],
        orders: npt.NDArray[np.int_],
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_], npt.NDArray[np.int_]]:
        """
        sorts all eigenvalues and eigenvectors for all orders
        """
        idx = eigenvalues.argsort()[::-1]
        eigenvalues = eigenvalues[idx]
        eigenvectors = eigenvectors[idx]
        orders = orders[idx]
        return eigenvalues, eigenvectors, orders

    def _create_Dm_matrix(  # noqa: N802
        self, m: int, emm: npt.NDArray[np.float_]
    ) -> npt.NDArray[np.float_]:
        """
        Syntax:
        Dm = _create_Dm_matrix(m, P)

        Input:
        m  =  order
        P(:,1)  =  Pl = Legendre Polynomials column vector for l = 0 : L-1
        P(:,2)  =  ell values vector

        Output:
        Dm = (L - m) square Slepian matrix for order m

        Description:
        This piece of code computes the Slepian matrix, Dm, for order m and all
        degrees, using the formulation given in "Spatiospectral Concentration on
        a Sphere" by F.J. Simons, F.A. Dahlen and M.A. Wieczorek.
        """
        Pl, ell = self._create_legendre_polynomials_table(emm)
        Dm = np.zeros((self.L - m, self.L - m))
        lvec = np.arange(m, self.L)

        Dm_ext, shm_ext = create_shared_memory_array(Dm)

        def func(chunk: list[int]) -> None:
            """
            calculate D matrix components for each chunk
            """
            Dm_int, shm_int = attach_to_shared_memory_block(Dm, shm_ext)

            # deal with chunk
            for i in chunk:
                logger.info(f"start ell: {i}")
                self._dm_matrix_helper(Dm_int, i, m, lvec, Pl, ell)
                logger.info(f"finish ell: {i}")

            free_shared_memory(shm_int)

        # split up L range to maximise effiency
        chunks = split_arr_into_chunks(self.L - m, NCPU)

        # initialise pool and apply function
        with ThreadPoolExecutor(max_workers=NCPU) as e:
            e.map(func, chunks)

        # retrieve from parallel function
        Dm = Dm_ext * (-1) ** m / 2

        # Free and release the shared memory block at the very end
        free_shared_memory(shm_ext)
        release_shared_memory(shm_ext)
        return Dm

    def _create_legendre_polynomials_table(
        self, emm: npt.NDArray[np.float_]
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.int_]]:
        """
        create Legendre polynomials table for matrix calculation
        """
        Plm = ssht.create_ylm(self.theta_max, 0, 2 * self.L).real.reshape(-1)
        ind = emm == 0
        ell = np.arange(2 * self.L)[np.newaxis]
        Pl = np.sqrt((4 * np.pi) / (2 * ell + 1)) * Plm[ind]
        return Pl, ell

    def _dm_matrix_helper(  # noqa: PLR0913
        self,
        Dm: npt.NDArray[np.float_],
        i: int,
        m: int,
        lvec: npt.NDArray[np.int_],
        Pl: npt.NDArray[np.float_],
        ell: npt.NDArray[np.int_],
    ) -> None:
        """
        used in both serial and parallel calculations
        """
        el = int(lvec[i])
        for j in range(i, self.L - m):
            p = int(lvec[j])
            c = 0
            for n in range(abs(el - p), el + p + 1):
                A = Pl[ell == n - 1] if n != 0 else 1
                c += (
                    self._wigner3j(el, n, p, 0, 0, 0)
                    * self._wigner3j(el, n, p, m, 0, -m)
                    * (A - Pl[ell == n + 1])
                )
            Dm[i, j] = (
                self._polar_gap_modification(el, p)
                * np.sqrt((2 * el + 1) * (2 * p + 1))
                * c
            )
            Dm[j, i] = Dm[i, j]

    @staticmethod
    def _wigner3j(  # noqa: PLR0913
        l1: int, l2: int, l3: int, m1: int, m2: int, m3: int
    ) -> float:
        """
        Syntax:
        s = _wigner3j (l1, l2, l3, m1, m2, m3)

        Input:
        l1  =  first degree in Wigner 3j symbol
        l2  =  second degree in Wigner 3j symbol
        l3  =  third degree in Wigner 3j symbol
        m1  =  first order in Wigner 3j symbol
        m2  =  second order in Wigner 3j symbol
        m3  =  third order in Wigner 3j symbol

        Output:
        s  =  Wigner 3j symbol for l1,m1; l2,m2; l3,m3

        Description:
        Computes Wigner 3j symbol using Racah formula
        """
        if (
            2 * l1 != np.floor(2 * l1)
            or 2 * l2 != np.floor(2 * l2)
            or 2 * l3 != np.floor(2 * l3)
            or 2 * m1 != np.floor(2 * m1)
            or 2 * m2 != np.floor(2 * m2)
            or 2 * m3 != np.floor(2 * m3)
        ):
            raise ValueError("Arguments must either be integer or half-integer!")

        if (
            m1 + m2 + m3 != 0
            or l3 < abs(l1 - l2)
            or l3 > (l1 + l2)
            or abs(m1) > abs(l1)
            or abs(m2) > abs(l2)
            or abs(m3) > abs(l3)
            or l1 + l2 + l3 != np.floor(l1 + l2 + l3)
        ):
            s = 0
        else:
            t1 = l2 - l3 - m1
            t2 = l1 - l3 + m2
            t3 = l1 + l2 - l3
            t4 = l1 - m1
            t5 = l2 + m2

            tmin = max(0, max(t1, t2))
            tmax = min(t3, min(t4, t5))

            # sum is over all those t for which the
            # following factorials have non-zero arguments
            s = sum(
                (-1) ** t
                / (
                    gp.factorial(t)
                    * gp.factorial(t - t1)
                    * gp.factorial(t - t2)
                    * gp.factorial(t3 - t)
                    * gp.factorial(t4 - t)
                    * gp.factorial(t5 - t)
                )
                for t in range(tmin, tmax + 1)
            )
            triangle_coefficient = (
                gp.factorial(t3)
                * gp.factorial(l1 - l2 + l3)
                * gp.factorial(-l1 + l2 + l3)
                / gp.factorial(l1 + l2 + l3 + 1)
            )

            s *= (
                np.float_power(-1, l1 - l2 - m3)
                * gp.sqrt(triangle_coefficient)
                * gp.sqrt(
                    gp.factorial(l1 + m1)
                    * gp.factorial(t4)
                    * gp.factorial(t5)
                    * gp.factorial(l2 - m2)
                    * gp.factorial(l3 + m3)
                    * gp.factorial(l3 - m3)
                )
            )
        return s

    def _polar_gap_modification(self, ell1: int, ell2: int) -> int:
        """
        eq 67 - Spherical Slepian functions and the polar gap in geodesy
        multiply by 1 + (-1)*(ell+ell')
        """
        return 1 + self.gap * (-1) ** (ell1 + ell2)

    def _clean_evals_and_evecs(
        self,
        eigenvalues: npt.NDArray[np.float_],
        gl: npt.NDArray[np.float_],
        emm: npt.NDArray[np.float_],
        m: int,
    ) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.complex_]]:
        """
        need eigenvalues and eigenvectors to be in a certain format
        """
        # Sort eigenvalues and eigenvectors in descending order of eigenvalues
        idx = eigenvalues.argsort()[::-1]
        eigenvalues = eigenvalues[idx]
        gl = gl[:, idx].conj()

        # put back in full D space for harmonic transform
        emm = emm[: self.L**2]
        ind = np.tile(emm == m, (self.L - abs(m), 1))
        eigenvectors = np.zeros((self.L - abs(m), self.L**2), dtype=np.complex_)
        eigenvectors[ind] = gl.T.flatten()

        # ensure first element of each eigenvector is positive
        eigenvectors *= np.where(eigenvectors[:, 0] < 0, -1, 1)[:, np.newaxis]

        # if -ve 'm' find orthogonal eigenvectors to +ve 'm' eigenvectors
        if m < 0:
            eigenvectors *= 1j

        return eigenvalues, eigenvectors

    @validator("order")
    def check_order(cls, v, values):
        if v is not None and (np.abs(v) >= values["L"]).any():
            raise ValueError(f"Order magnitude should be less than {values['L']}")
        return v

    @validator("theta_max")
    def check_theta_max(cls, v):
        if v == 0:
            raise ValueError("theta_max cannot be zero")
        return v



import numpy as np
import pyssht as ssht
from numpy import typing as npt

from sleplet import logger


def apply_gaussian_smoothing(
    flm: npt.NDArray[np.complex_], L: int, smoothing_factor: int
) -> npt.NDArray[np.complex_]:
    """
    applies Gaussian smoothing to the given signal

    s_lm = exp(-ell^2 sigma^2)
    s(omega) = exp(-theta^2 / sigma^2)
    FWHM = 2 * sqrt(ln2) * sigma
    """
    sigma = np.pi / (smoothing_factor * L)
    fwhm = 2 * np.sqrt(np.log(2)) * sigma
    logger.info(f"FWHM = {np.rad2deg(fwhm):.2f}degrees")
    return ssht.gaussian_smoothing(flm, L, sigma)



from argparse import ArgumentParser
from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt

from sleplet.functions.flm.axisymmetric_wavelet_coefficients_south_america import (
    AxisymmetricWaveletCoefficientsSouthAmerica,
)
from sleplet.functions.fp.slepian_wavelet_coefficients_south_america import (
    SlepianWaveletCoefficientsSouthAmerica,
)
from sleplet.utils.plot_methods import save_plot
from sleplet.utils.region import Region

fig_path = Path(__file__).resolve().parents[3] / "src" / "sleplet" / "figures"
sns.set(context="paper")

B = 3
J_MIN = 2
L = 128
STEP = 0.01


def _plot_slepian_coefficients() -> int:
    """
    plot the Slepian wavelet coefficients for the South America region
    """
    # initialise wavelet coefficients
    region = Region(mask_name="south_america")
    swc = SlepianWaveletCoefficientsSouthAmerica(L, B=B, j_min=J_MIN, region=region)

    # find sorted coefficients
    w_p = np.sort(np.abs(swc.wavelet_coefficients), axis=1)[:, ::-1]

    # perform plot
    plt.plot(w_p[0], label=r"$|W^{{\phi}}_{p}|$")

    for j, w_j in enumerate(w_p[1:]):
        plt.plot(w_j, label=rf"$|W^{{\phi^{j+J_MIN}}}_{{p}}|$")

    return swc.slepian.N


def _plot_axisymmetric_coefficients(shannon: int) -> None:
    """
    plot the axisymmetric wavelet coefficients for the South America region
    """
    # initialise wavelet coefficients
    awc = AxisymmetricWaveletCoefficientsSouthAmerica(L, B=B, j_min=J_MIN)

    # find sorted coefficients
    w_lm = np.sort(np.abs(awc.wavelet_coefficients), axis=1)[:, ::-1]

    # perform plot
    plt.plot(w_lm[0, :shannon], "--", label=r"$|W^{{\phi}}_{\ell m}|$")

    for j, w_j in enumerate(w_lm[1:]):
        plt.plot(w_j[:shannon], "--", label=rf"$|W^{{\phi^{j+J_MIN}}}_{{\ell m}}|$")


def main(*, limit: bool) -> None:
    """
    Plot a comparison of the absolute values of the wavelet coefficients
    compared to the Slepian coefficients. Expect the Slepian coefficients to
    decay faster than the wavelet coefficients.
    """
    shannon = _plot_slepian_coefficients()
    _plot_axisymmetric_coefficients(shannon)
    plt.xlabel("coefficient index")
    plt.legend()
    filename = f"south_america_sparsity_wavelet_coefficients_comparison_L{L}"
    if limit:
        plt.xlim(right=300)
        plt.ylim(top=50)
        filename += "_limit"
    save_plot(fig_path, filename)


if __name__ == "__main__":
    parser = ArgumentParser(description="South America sparsity")
    parser.add_argument(
        "--limit",
        "-l",
        action="store_true",
        help="flag which limits the region of the plot",
    )
    args = parser.parse_args()
    main(limit=args.limit)



import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

from sleplet.functions.flm.earth import Earth
from sleplet.functions.flm.harmonic_gaussian import HarmonicGaussian
from sleplet.functions.flm.identity import Identity
from sleplet.functions.fp.slepian_identity import SlepianIdentity
from sleplet.utils.slepian_methods import slepian_forward

L = 128


def test_earth_identity_convolution() -> None:
    """
    test to ensure that the convolving with the
    identity function doesn't change the map
    """
    f = Earth(L)
    g = Identity(L)
    flm = f.coefficients
    flm_conv = f.convolve(flm, g.coefficients)
    assert_array_equal(flm, flm_conv)


def test_earth_harmonic_gaussian_convolution() -> None:
    """
    test to ensure that convolving the Earth with the harmonic
    Gausian does not change significantly change the map
    """
    f = Earth(L)
    g = HarmonicGaussian(L)
    flm = f.coefficients
    flm_conv = f.convolve(flm, g.coefficients)
    assert_allclose(np.abs(flm_conv - flm).mean(), 0, atol=7)


def test_south_america_slepian_identity_convolution(
    slepian_arbitrary, south_america_arbitrary
) -> None:
    """
    test to ensure that the convolving with the Slepian
    identity function doesn't change the map in Slepian space
    """
    f_p = slepian_forward(
        slepian_arbitrary.L, slepian_arbitrary, flm=south_america_arbitrary.coefficients
    )
    g = SlepianIdentity(slepian_arbitrary.L, region=slepian_arbitrary.region)
    fp_conv = south_america_arbitrary.convolve(
        f_p, g.coefficients, shannon=slepian_arbitrary.N
    )
    assert_array_equal(f_p, fp_conv)



import os
from pathlib import Path

import pooch

from sleplet import logger
from sleplet.utils.vars import ZENODO_DATA_DOI

_data_path = Path(__file__).resolve().parent

POOCH = pooch.create(
    path=pooch.os_cache("sleplet"),
    # Use the figshare DOI
    base_url=f"doi:{ZENODO_DATA_DOI}/",
    registry=None,
)
POOCH.load_registry_from_doi()


def find_on_pooch_then_local(filename: str) -> os.PathLike | None:
    """
    find a file on POOCH first and if not look in data folder
    """
    if filename in POOCH.registry:
        logger.info(f"Found {filename} at https://doi.org/{ZENODO_DATA_DOI}")
        return POOCH.fetch(filename, progressbar=True)
    elif (_data_path / filename).exists():
        logger.info(f"Found {filename} at {_data_path / filename}")
        return _data_path / filename
    else:
        logger.info(f"No {filename} found, calculating...")
        return None



import numpy as np
from multiprocess.shared_memory import SharedMemory
from numpy import typing as npt


def split_arr_into_chunks(arr_max: int, ncpu: int) -> list[npt.NDArray[np.int_]]:
    """
    split L into a list of arrays for parallelism
    """
    arr = np.arange(arr_max)
    size = len(arr)
    arr[size // 2 : size] = arr[size // 2 : size][::-1]
    return [np.sort(arr[i::ncpu]) for i in range(ncpu)]


def create_shared_memory_array(
    array: npt.NDArray[np.float_],
) -> tuple[npt.NDArray[np.float_], SharedMemory]:
    """
    creates a shared memory array to be used in a parallel function
    """
    ext_shared_memory = SharedMemory(create=True, size=array.nbytes)
    array_ext: npt.NDArray[np.float_] = np.ndarray(
        array.shape, dtype=array.dtype, buffer=ext_shared_memory.buf
    )
    return array_ext, ext_shared_memory


def attach_to_shared_memory_block(
    array: npt.NDArray[np.float_], ext_shared_memory: SharedMemory
) -> tuple[npt.NDArray[np.float_], SharedMemory]:
    """
    used within the parallel function to attach an array to the shared memory
    """
    int_shared_memory = SharedMemory(name=ext_shared_memory.name)
    array_int: npt.NDArray[np.float_] = np.ndarray(
        array.shape, dtype=array.dtype, buffer=int_shared_memory.buf
    )
    return array_int, int_shared_memory


def free_shared_memory(*shared_memory: SharedMemory) -> None:
    """
    closes the shared memory object
    """
    for shm in shared_memory:
        shm.close()


def release_shared_memory(*shared_memory: SharedMemory) -> None:
    """
    releases the shared memory object
    """
    for shm in shared_memory:
        shm.unlink()



from pathlib import Path

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt

from sleplet.slepian.slepian_region.slepian_arbitrary import SlepianArbitrary
from sleplet.utils.plot_methods import save_plot

fig_path = Path(__file__).resolve().parents[3] / "src" / "sleplet" / "figures"
sns.set(context="paper")

L = 128


def main() -> None:
    """
    plots the tiling of the Slepian line
    """
    slepian = SlepianArbitrary(L, "africa")
    p_range = np.arange(0, L**2)
    plt.semilogx(p_range, slepian.eigenvalues, "k.")
    plt.axvline(slepian.N, c="k", ls="--", alpha=0.8)
    plt.annotate(
        f"N={slepian.N}",
        xy=(slepian.N, 1),
        xytext=(19, 3),
        ha="center",
        textcoords="offset points",
        annotation_clip=False,
    )
    ticks = 2 ** np.arange(np.log2(L**2) + 1, dtype=int)
    plt.xticks(ticks, ticks)
    plt.xlabel(r"$p$")
    plt.ylabel(r"$\mu$")
    save_plot(fig_path, f"africa_eigenvalues_L{L}")


if __name__ == "__main__":
    main()



from argparse import ArgumentParser
from pathlib import Path

import cmocean
import numpy as np

from sleplet.meshes.classes.mesh import Mesh
from sleplet.plotting.create_plot_mesh import Plot
from sleplet.scripts.plotting_on_mesh import valid_meshes
from sleplet.utils.class_lists import MESHES

fig_path = Path(__file__).resolve().parents[2] / "src" / "sleplet" / "figures"


def main(mesh_name: str) -> None:
    """
    plots the tiling of the Slepian line
    """
    # initialise mesh and Slepian mesh
    mesh = Mesh(mesh_name)

    # create region masking
    field = np.zeros(mesh.vertices.shape[0])
    field[mesh.region] = 1

    name = f"{mesh_name}_region"
    Plot(mesh, name, field, colour=cmocean.cm.haline, region=True).execute()


if __name__ == "__main__":
    parser = ArgumentParser(description="mesh tiling")
    parser.add_argument(
        "function",
        type=valid_meshes,
        choices=MESHES,
        help="mesh to plot",
    )
    args = parser.parse_args()
    main(args.function)



from numpy.testing import assert_array_equal, assert_array_less, assert_raises

from sleplet.functions.flm.axisymmetric_wavelets import AxisymmetricWavelets
from sleplet.functions.flm.earth import Earth
from sleplet.utils.denoising import denoising_axisym

B = 2
J_MIN = 0
L = 128
N_SIGMA = 3
SNR_IN = 10


def test_denoising_earth_axisymmetric_wavelets() -> None:
    """
    tests that hard thresholding improves the SNR over the map
    """
    fun = Earth(L)
    fun_noised = Earth(L, noise=SNR_IN)
    aw = AxisymmetricWavelets(L, B=B, j_min=J_MIN)
    _, noised_snr, denoised_snr = denoising_axisym(fun, fun_noised, aw, SNR_IN, N_SIGMA)
    assert isinstance(noised_snr, float)
    assert_array_less(noised_snr, denoised_snr)


def test_adding_noise_changes_flm() -> None:
    """
    tests the addition of Gaussian noise changes the coefficients
    """
    earth = Earth(L)
    earth_noised = Earth(L, noise=SNR_IN)
    assert_raises(
        AssertionError,
        assert_array_equal,
        earth.coefficients,
        earth_noised.coefficients,
    )
